/* $Id: tcp_bw_exp.c,v 1.2 2000/03/28 15:04:47 susan Exp $ */

#include "config.h"
#include <sys/types.h>     /* system type definitions */
#include <stdlib.h>        /* free() malloc()  */
#include <netinet/in.h>    /* sometimes required for #include <arpa/inet.h> */
#include <sys/socket.h>    /* setsockopt() */
#include <arpa/inet.h>     /* inet_ntoa() */
#include <string.h>        /* strerror() */
#include <errno.h>         /* errno */
#include <sys/time.h>      /* struct timeval */
#include <unistd.h>        /* select() on some systems */
#include "protocol.h"      /* Socket functions. */
#include "host_protocol.h" /* HostDImage() */
#include "diagnostic.h"    /* FAIL() WARN() */
#include "osutil.h"        /* CurrentTime() MicroTime() SetRealTimer() */
#include "strutil.h"       /* SAFESTRNCPY() vstrncpy() */
#include "messages.h"      /* {Recv,Send}Message() */
#include "tcp_bw_exp.h"


#define TCP_HANDSHAKE 411
typedef struct {
  IPAddress ipAddr; /* Reserved for future use. */
  unsigned short portNum;
} TcpHandshake;
static const DataDescriptor tcpHandshakeDescriptor[] =
  {SIMPLE_MEMBER(UNSIGNED_INT_TYPE, 1, offsetof(TcpHandshake, ipAddr)),
   SIMPLE_MEMBER(UNSIGNED_SHORT_TYPE, 1, offsetof(TcpHandshake, portNum))};
#define tcpHandshakeDescriptorLength 2


int
InitiateTcpExp(IPAddress machine,
               unsigned short port,
               MessageType requestMessage,
               const TcpBwCtrl *sizes,
               double timeOut,
               double *results) {

  char *expData;
  int msgs;
  struct sockaddr_in server;
  Socket serverSock;
  long startTime;
  long stopTime;
  TcpHandshake handshake;
  int sentThisCall;
  int sentThisMessage;
  int totalSent;

  if(!CallAddr(machine, port, &serverSock, timeOut)) {
    FAIL2("InitiateTcpExp: attempt to contact %s:%d failed\n",
          IPAddressMachine(machine), port);
  }

  LOG5("TCP(%d, %d, %d) to %s:%d\n",
       sizes->expSize, sizes->bufferSize, sizes->msgSize,
       IPAddressMachine(machine), port);

  if(!SendMessageAndData(serverSock,
                         requestMessage,
                         sizes,
                         tcpBwCtrlDescriptor,
                         tcpBwCtrlDescriptorLength,
                         PktTimeOut(serverSock))) {
    DROP_SOCKET(&serverSock);
    FAIL("InitiateTcpExp: request send failed\n");
  }

  /*
  ** Note: the server has to set up a new listening port before it can respond,
  ** so we use a generous time-out value here.
  */
  if(!RecvMessageAndData(serverSock,
                         TCP_HANDSHAKE,
                         &handshake,
                         tcpHandshakeDescriptor,
                         tcpHandshakeDescriptorLength,
                         MAXPKTTIMEOUT)) {
    DROP_SOCKET(&serverSock);
    FAIL("InitiateTcpExp: receive of handshake failed\n");
  }

  /* Reestablish the connection on the port contained in the handshake. */
  DROP_SOCKET(&serverSock);
  if((serverSock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    FAIL("InitiateTcpExp: socket() failed\n");
  }

  if(setsockopt(serverSock,
                SOL_SOCKET,
                SO_SNDBUF,
                (const char *)&sizes->bufferSize,
                sizeof(sizes->bufferSize)) < 0) {
    WARN1("InitiateTcpExp: set send buffer size %d failed\n",
          sizes->bufferSize);
  }

  memset(&server, 0, sizeof(server));
  server.sin_addr.s_addr = machine;
  server.sin_family = AF_INET;
  server.sin_port = htons(handshake.portNum);

  SetRealTimer(MINCONNTIMEOUT);
  if(connect(serverSock, (struct sockaddr *)&server, sizeof(server)) < 0) {
    RESETREALTIMER;
    DROP_SOCKET(&serverSock);
    FAIL2("InitiateTcpExp: connect to experiment socket %s:%d failed\n",
          IPAddressMachine(machine), handshake.portNum);
  }
  RESETREALTIMER;

  /* Initialize enough data to cover the bandwidth experiment. */
  if((expData = (char *)malloc((size_t)sizes->msgSize)) == NULL) {
    DROP_SOCKET(&serverSock);
    FAIL1("InitiateTcpExp: malloc %d failed\n", sizes->msgSize);
  }
  memset(expData, 0, sizes->msgSize);

  /* Latency experiment... */
  SetRealTimer((int)PktTimeOut(serverSock));
  startTime = MicroTime();

  if(send(serverSock, expData, 1, 0) <= 0) {
    RESETREALTIMER;
    free(expData);
    DROP_SOCKET(&serverSock);
    FAIL("InitiateTcpExp: latency send failed\n");
  }

  if(recv(serverSock, expData, 1, 0) <= 0) {
    RESETREALTIMER;
    free(expData);
    DROP_SOCKET(&serverSock);
    FAIL("InitiateTcpExp: latency receive failed\n");
  }

  stopTime = MicroTime();
  RESETREALTIMER;
  results[1] = ((double)(stopTime - startTime)) / 1000.0;

  /* ... then bandwidth. */
  msgs = sizes->expSize / sizes->msgSize + 1;
  SetRealTimer((int)PktTimeOut(serverSock) * msgs);
  startTime = MicroTime();

  for(totalSent = 0; totalSent < sizes->expSize; totalSent += sentThisMessage) {
    for(sentThisMessage = 0;
        sentThisMessage < sizes->msgSize;
        sentThisMessage += sentThisCall) {
      sentThisCall =
        send(serverSock, expData, sizes->msgSize - sentThisMessage, 0);
      if(sentThisCall <= 0) {
        RESETREALTIMER;
        free(expData);
        DROP_SOCKET(&serverSock);
        FAIL1("InitiateTcpExp: data transfer failed after %d bytes\n",
              totalSent + sentThisMessage);
      }
    }
  }

  RESETREALTIMER;
  free(expData);

  /*
  ** Wait for the receiver to close the connection so that we're sure that all
  ** the data has been sent and received, rather than locally buffered.
  */
  if(!CloseSocket(&serverSock, PktTimeOut(serverSock))) {
    DROP_SOCKET(&serverSock);
    FAIL("InitiateTcpExp: no close received from server\n");
  }

  stopTime = MicroTime();
  results[0] = ( ((double)sizes->expSize * 8.0) /
                 ((double)(stopTime - startTime) / 1000000.0) ) / 1000000.0;

  return 1;

}


int
TerminateTcpExp(Socket sender) {

  struct sockaddr_in client;
  SOCKLEN_T clientSize = sizeof(client);
  Socket clientSock;
  char *expData;
  TcpBwCtrl sizes;
  int msgs;
  Socket earSock;
  fd_set earFD;
  TcpHandshake handshake;
  int recvThisCall;
  int recvThisMessage;
  struct sockaddr_in server;
  SOCKLEN_T serverSize = sizeof(server);
  struct timeval timeOut;
  int totalRecv;

  if(!RecvData(sender,
               &sizes,
               tcpBwCtrlDescriptor,
               tcpBwCtrlDescriptorLength,
               PktTimeOut(sender))) {
    FAIL("TerminateTcpExp: configuration receive failed\n");
  }

  LOG3("Servicing TCP(%d, %d, %d)\n",
       sizes.expSize, sizes.bufferSize, sizes.msgSize);

  /* Establish a listening port with the experiment buffer size. */
  earSock = socket(AF_INET, SOCK_STREAM, 0);
  if(setsockopt(earSock,
                SOL_SOCKET,
                SO_RCVBUF,
                (const char *)&sizes.bufferSize,
                sizeof(sizes.bufferSize)) < 0) {
    WARN2("TerminateTcpExp: set buff size failed %d:%s\n",
          errno, strerror(errno));
  }

  memset(&server, 0, sizeof(server));
  server.sin_addr.s_addr = INADDR_ANY;
  server.sin_family = AF_INET;
  server.sin_port = 0;

  if((bind(earSock, (struct sockaddr *)&server, sizeof(server)) < 0) ||
     (listen(earSock, 1) < 0) ||
     (getsockname(earSock, (struct sockaddr *)&server, &serverSize) < 0)) {
    DROP_SOCKET(&earSock);
    FAIL("TerminateTcpExp: bind/listen failed.\n");
  }

  /* Tell the client what port to contact, then wait for them to respond. */
  handshake.ipAddr = (IPAddress)0;
  handshake.portNum = ntohs(server.sin_port);

  if(!SendMessageAndData(sender,
                         TCP_HANDSHAKE,
                         &handshake,
                         tcpHandshakeDescriptor,
                         tcpHandshakeDescriptorLength,
                         PktTimeOut(sender))) {
    DROP_SOCKET(&earSock);
    FAIL1("TerminateTcpExp: send of reply failed on %d\n", sender);
  }

  /* Give the client 10 seconds to connect. */
  FD_ZERO(&earFD);
  FD_SET(earSock, &earFD);
  timeOut.tv_sec = 10;
  timeOut.tv_usec = 0;

  if((select(earSock + 1, &earFD, NULL, NULL, &timeOut) < 1) ||
     ((clientSock = accept(earSock,
                           (struct sockaddr *)&client,
                           &clientSize)) < 0)) {
    DROP_SOCKET(&earSock);
    FAIL("TerminateTcpExp: accept failed\n");
  }

  /* Initialize enough data to cover the bandwidth experiment. */
  if((expData = (char *)malloc(sizes.msgSize)) == NULL) {
    DROP_SOCKET(&clientSock);
    DROP_SOCKET(&earSock);
    FAIL1("TerminateTcpExp: malloc %d failed\n", sizes.msgSize);
  }
  memset(expData, 0, sizes.msgSize);

  /* Latency experiment ... */
  SetRealTimer((int)PktTimeOut(clientSock));

  if(recv(clientSock, expData, 1, 0) <= 0) {
    RESETREALTIMER;
    free(expData);
    DROP_SOCKET(&clientSock);
    DROP_SOCKET(&earSock);
    FAIL("TerminateTcpExp: latency recv failed\n");
  }

  if(send(clientSock, expData, 1, 0) <= 0) {
    RESETREALTIMER;
    free(expData);
    DROP_SOCKET(&clientSock);
    DROP_SOCKET(&earSock);
    FAIL("TerminateTcpExp: latency send failed\n");
  }

  /* ... then bandwidth. */
  msgs = sizes.expSize / sizes.msgSize + 2;
  SetRealTimer((int)PktTimeOut(clientSock) * msgs);

  for(totalRecv = 0; totalRecv < sizes.expSize; totalRecv += recvThisMessage) {
    for(recvThisMessage = 0;
        recvThisMessage < sizes.msgSize;
        recvThisMessage += recvThisCall) {
      recvThisCall =
        recv(clientSock, expData, sizes.msgSize - recvThisMessage, 0);
      if(recvThisCall <= 0) {
        RESETREALTIMER;
        DROP_SOCKET(&clientSock);
        DROP_SOCKET(&earSock);
        free(expData);
        FAIL1("TerminateTcpExp: connection failed after %d bytes\n",
              totalRecv + recvThisMessage);
      }
    }
  }

  RESETREALTIMER;
  DROP_SOCKET(&clientSock);
  DROP_SOCKET(&earSock);
  free(expData);

  return 1;

}
