/*****************************************************************/
/*      agent_init.c                                             */
/*      Henri Casanova, Dorian Arnold                            */
/*---------------------------------------------------------------*/
/*  agentInit()                                                 */
/*****************************************************************/

#include <signal.h>
#include <fcntl.h>
#include <netdb.h>
#include <sys/stat.h>
#include "core.h"
#include "agentglobal.h"
#include "agentinit.h"
#include "nwsutils.h"

#ifdef NWS
#include "nws_api.h"
#endif

#include "agentwakeup.h"
#include "exchangemappings.h"
#include "netsolvesignals.h"
#include <unistd.h>
extern int errno;

/*
 *  agentInit()
 *
 *  Performs all the initializations required by the agent
 */
int agent_init(int argc,char **argv)
{
  /* initialize the NETSOLVE_ROOT path */
  global.netsolve_root_path=getNetSolveRoot();
  if (global.netsolve_root_path == NULL)
  {
    fprintf(stderr,"The $NETSOLVE_ROOT variable should be set\n");
    return -1;
  }  

  set_defaults();
  check_args(argc, argv);

  /* Initialize the NSLinked Lists */
  global.problems = newNSLinkedList();
  global.agents   = newNSLinkedList();
  global.servers  = newNSLinkedList();
  global.mappings = newNSLinkedList();


#ifndef NO_OUTPUT
  fprintf(stderr,"Initializing Network connections...\n");
#endif
  /* Initialize my_self */
  if (initMySelf() == -1)
    return -1;

  /* Signals */
  if (setSignals() == -1)
    return -1;

#ifdef NWS
  if (setNWS()==-1)
  return -1;
#endif

  /* Register to the NetSolve system */
  if (registerToNetSolve() == -1)
    return -1;

#ifdef HBMDC
#ifndef NO_OUTPUT
  fprintf(stderr,"Globus HBM Data Collector...\n");
#endif
  /* Running the HBM */
  if (runHBMDataCollector() == -1)
    return -1;
#endif

#ifdef HBM
#ifndef NO_OUTPUT
  fprintf(stderr,"Globus HBM NetSolve Proxy...\n");
#endif
  /* Running the HBM */
  if (runHBMProxy() == -1)
    return -1;
#endif

  if(daemon_init() < 0 ){
    return -1;
  }

  fflush(stderr);
  fflush(stdout);
  return 0;
}

/*
 * initMySelf()
 *
 * Performs the initialization of the host descriptor
 * the IP port and the listening socket
 */
int initMySelf()
{
  struct hostent *hp;
  char buffer[256];
  
  if(!global.my_self->host_desc->hostname) {
    /* Getting my host name and IPaddr */
    if (gethostname(buffer,256) == -1)
      {
	perror("gethostname()");
	ns_errno = NetSolveSystemError;
	return -1;
      }
    if ((hp = gethostbyname(buffer)) == NULL)
      {
	perror("gethostbyname():");
	ns_errno = NetSolveSystemError;
	return -1;
      }
    BCOPY(hp->h_addr_list[0],
	  &(global.my_self->host_desc->IPaddr),sizeof(NS_IPaddr_type));
    global.my_self->host_desc->hostname = strdup(hp->h_name);
  }

  fprintf(stderr, "Hostname set to \"%s\"\n", global.my_self->host_desc->hostname);
  global.my_self->port=AGENT_PORT;
  global.sock=establishSocket(&(global.my_self->port), 1, global.my_self->host_desc->hostname);

  if (global.sock < 0)
  {
    fprintf(stderr,"Impossible to bind to port\n");
    return -1; 
  }

#ifndef NO_OUTPUT
  fprintf(stderr,"Binding to port %d...\n",global.my_self->port);
#endif

  global.my_self->host_desc->host_status = HOST_UP;
  global.my_self->host_desc->data_format = getArch();
  if (global.my_self->host_desc->data_format == -1)
  {
    fprintf(stderr,"Impossible to determine my architecture\n");
    ns_errno = NetSolveInternalError;
    return -1;
  }
  return 0;
}

/*
 * setMaster()
 *
 * Build the agent descriptor
 */
int setMaster(char *mastername)
{
  struct hostent *hp;
  NS_AgentDesc *agentdesc;

  /* Get the full name and IPaddr of the agent */
  hp = gethostbyname(mastername);
  if (hp == NULL)
  {
    perror("gethostbyname():");
    ns_errno = NetSolveSystemError;
    return -1;
  } 
  agentdesc = newAgentDesc();
  agentdesc->host_desc = newHostDesc();
  agentdesc->port = AGENT_PORT;
  agentdesc->host_desc->hostname= 
      strdup(hp->h_name);
  BCOPY(hp->h_addr_list[0],
       &(agentdesc->host_desc->IPaddr),
       sizeof(NS_IPaddr_type));

  addAgent(agentdesc);
  global.master = agentdesc;
  return 0;
}

/*
 * registerToNetSolve
 *
 * contacts the master agent and try to get registered
 */
int registerToNetSolve()
{
  int sock;
  NS_Communicator *comm;
  NS_AgentDesc *ad;
  int nb_problems;
  NS_ServerDesc *sd;
  int nb_servers,nb_agents,i;
  NS_LinkItem *link;
  NS_ProblemDesc *pd;
  int tag;

  /* Nothing to do if I am stand-alone */
  if (global.master == NULL)
    return 0;

#ifndef NO_OUTPUT
  fprintf(stderr, "Registering with agent %s\n",
                            global.master->host_desc->hostname);
#endif

  sock = connectToSocket(global.master->host_desc->hostname,
                           global.master->host_desc->IPaddr,
                           global.master->port);
  if (sock <0)
  {
    fprintf(stderr,"Impossible to contact agent on '%s'\n",
                   global.master->host_desc->hostname);
    return -1;
  }

  comm=initTransaction(sock,DATA_XDR);
  if (comm == NULL)
  {
    netsolvePerror(__FILE__,__LINE__);
    endTransaction(comm);
    return -1;
  }
  if (sendInt(comm,NS_PROT_AG_REGISTER) == -1)
  {
    netsolvePerror(__FILE__,__LINE__);
    endTransaction(comm);
    return -1;
  }
  if (sendAgentDesc(comm,global.my_self) == -1)
  {
    netsolvePerror(__FILE__,__LINE__);
    endTransaction(comm);
    return -1;
  }

  if (recvInt(comm,&tag) == -1)
  {
    netsolvePerror(__FILE__,__LINE__);
    endTransaction(comm);
    return -1;
  }
  if (tag == NS_PROT_REGISTRATION_REFUSED)
  {
    fprintf(stderr,"Rejected by the agent\n");
    endTransaction(comm);
    return -1;
  }
  if (tag != NS_PROT_REGISTRATION_ACCEPTED)
  {
    ns_errno = NetSolveProtocolError;
    netsolvePerror(__FILE__,__LINE__);
    endTransaction(comm);
  }
#ifdef VIEW
  fprintf(stderr,"Registration accepted\n");
#endif

  if (recvInt(comm,&nb_agents) == -1)
  {
    netsolvePerror(__FILE__,__LINE__);
    endTransaction(comm);
    return -1;
  }
  for (i=0;i<nb_agents;i++)
  {
    ad = recvAgentDesc(comm);
    if (ad == NULL)
    {
      netsolvePerror(__FILE__,__LINE__);
      endTransaction(comm);
      return -1;
    }
    if (ad->host_desc->IPaddr == global.my_self->host_desc->IPaddr)
      freeAgentDesc(ad);
    else
      addAgent(ad);
  }
  if (recvInt(comm,&nb_servers) == -1)
  {
    netsolvePerror(__FILE__,__LINE__);
    endTransaction(comm);
    return -1;
  }
  for (i=0;i<nb_servers;i++)
  {
    sd = recvServerDesc(comm);
    if (sd == NULL)
    {
      netsolvePerror(__FILE__,__LINE__);
      endTransaction(comm);
      return -1;
    }
    addServer(sd);
  }
  
  /* Recv all the problems */
  if (recvInt(comm,&nb_problems) == -1)
  {
    netsolvePerror(__FILE__,__LINE__);
    endTransaction(comm);
    return -1;
  }
  for (i=0;i<nb_problems;i++)
  {
    pd = recvProblemDesc(comm);
    if (pd == NULL)
    {
      netsolvePerror(__FILE__,__LINE__);
      endTransaction(comm);
      return -1;
    }
    addProblem(pd);
  }
 

  /* Recv the mappings */
  if (recvMappings(comm) == -1)
  {
    netsolvePerror(__FILE__,__LINE__);
    endTransaction(comm);
    return -1;
  }

  /* Notify my existence to the servers */
  link = global.servers->start;
  while(link != NULL)
  {
    sd = (NS_ServerDesc *)(link->content);
    fprintf(stderr,"Reporting to %s\n",sd->host_desc->hostname);
    reportToServer(sd);
    link = link->next;
  }

  endTransaction(comm);
  return 0; 
}

#if 0
#ifdef HBM
/*
 * runHBMProxy()
 *
 * Runs the Heart Beat Proxy 
 */
int runHBMProxy()
{
  char commandline[256];
  char path[256];
  int pid;
  char *arch;

  global.hbmproxy_pid = fork();
  if (global.hbmproxy_pid == -1)
  {
    ns_errno = NetSolveSystemError;
    netsolvePerror(__FILE__,__LINE__);
    return -1;
  }

  if (global.hbmproxy_pid != 0)
    return 0;

  arch = getNetSolveArch();
  sprintf(path,"%s/bin/%s/",global.netsolve_root_path,arch);
  free(arch);
  if (chdir(path)== -1)
  {
    fprintf(stderr,"Warning: Unable to start the HBM proxy.\n");
    perror("chdir");
    exit(-1);
  }
  if (execl("./hbmproxy","hbmproxy",(char *)0) == -1)
  {
    fprintf(stderr,"Warning: Unable to start the HBM proxy.\n");
    perror("execl");
    exit(-1);
  }
  exit(1);
}
#endif
#endif


#ifdef HBMDC
/*
 * runHBMDataCollector()
 *
 * Runs the Heart Beat Data Collector
 */
int runHBMDataCollector()
{
  char commandline[256];
  char path[256];
  int pid;
  char *arch;

  /* clean up the HBM files if any */
  sprintf(commandline,"rm -rf /tmp/hbmdc*");
  system(commandline);

  global.hbmdc_pid = fork();
  if (global.hbmdc_pid == -1)
  {
    ns_errno = NetSolveSystemError;
    netsolvePerror(__FILE__,__LINE__);
    return -1;
  }

  if (global.hbmdc_pid != 0)
    return 0;  
 
  arch = getNetSolveArch();
  if (execl(HBMDC,"hbmdc",(char *)0) == -1)
  {
    fprintf(stderr,"Warning: Unable to start Globus HBM Data Collector.\n");
    perror("execl");
    exit(-1);
  }
  exit(1);
}
#endif

/*
 * reportToServer()
 *
 * report my existence to all the servers that I have 
 * received from my master agent.
 */
void reportToServer(NS_ServerDesc *sd)
{
  int sock;
  NS_Communicator *comm;

  sock = connectToSocket(sd->host_desc->hostname,
                         sd->host_desc->IPaddr,
                         sd->port);
  if (sock < 0)
  {
#ifndef NO_OUTPUT
    fprintf(stderr,"WARNING:Impossible to contact server on '%s'\n",
            sd->host_desc->hostname);
    return;
#endif
  }
  comm = initTransaction(sock,DATA_XDR);
  if (comm == NULL)
  {
    netsolvePerror(__FILE__,__LINE__);
    close(sock);
    return;
  }
  if (sendInt(comm,NS_PROT_NEW_AGENT) == -1)
  {
    netsolvePerror(__FILE__,__LINE__);
    endTransaction(comm);
    return;
  }
  if (sendAgentDesc(comm,global.my_self) == -1)
  {
    netsolvePerror(__FILE__,__LINE__);
    endTransaction(comm);
    return;
  }
  endTransaction(comm);
  return;
}

void print_usage(char *command){
  fprintf(stderr,"Usage: %s [ -a agent_name ] [ -l logfile ]\n"
                  , command);
  exit(-1);
}

int check_args(int argc, char **argv){
  int i, retval=0;

  for(i=1; i<argc; i++){
    if(strcmp(argv[i], "-a") == 0){
      if(i == argc-1){   /* last argument? */
        print_usage(argv[0]);
        return -1;
      }
      else{
        retval = setMaster(argv[i+1]);
        i++;
        if(retval < 0)
          return -1;
      }
    }

    else if(strcmp(argv[i], "-p") == 0) {
      if(i == argc-1) {
	print_usage(argv[0]);
	return -1;
      } else {
	global.my_self->port = strtoul(argv[i+1], 0, 0);
	i++;
	if(global.my_self->port == ULONG_MAX) { /* strtoul() returns ULONG_MAX on error */
	  fprintf(stderr, "Hey, port number given must be a valid port number!\n");
	  return -1;
	}
      }
    }

    else if(strcmp(argv[i], "-h") == 0) {
      if(i == argc-1) {
	print_usage(argv[0]);
	return -1;
      } else {
	global.my_self->host_desc->hostname = strdup(argv[i+1]);
	i++;
      }
    }

    else if(strcmp(argv[i], "-l") == 0){
      if(i == argc-1){   /* last argument? */
        print_usage(argv[0]);
        return -1;
      }
      else{
        global.log_file = (char *)strdup(argv[i+1]);
        i++;
      }
    }

    else{
      print_usage(argv[0]);
      return -1;
    }
  }

  return 0;
}

int set_defaults(){

  /*** set logfile to $NETSOLVE_ROOT/AGENT_LOGFILE **/
  global.log_file=(char *)calloc(strlen(global.netsolve_root_path) +
                                   strlen(AGENT_LOGFILE) + 2, sizeof(char));
  strcpy(global.log_file, global.netsolve_root_path);
  strcat(global.log_file, "/");
  strcat(global.log_file, AGENT_LOGFILE);

  global.master = NULL;

  global.my_self = newAgentDesc();
  global.my_self->host_desc = newHostDesc();

  return 0;
}

int daemon_init(){
  pid_t pid;
  int fd;

  pid = fork();

  if(pid < 0){
    perror("fork()");
    return -1;
  }

  if(pid != 0) /* parent exits */
    exit(0);

  setsid();   /* become session leader */

  pid = fork();
 
  if(pid < 0){
    perror("fork()");
    return -1;
  }
 
  if(pid != 0) /* 1st child exits */
    exit(0);

  chdir(global.netsolve_root_path);
  umask(0);

  fd = open("/dev/null", O_RDWR|O_CREAT);
  if(fd < 0){
    perror("open()");
    return -1;
  }
  close(0);
  if(dup2(fd, 0) < 0){
    perror("dup2()");
    return -1;
  }
  close(fd);

#ifndef NO_OUTPUT
  fprintf(stderr, "\n-- Agent giving up terminal control! --\n\n");
  fprintf(stderr, "All further terminal output will be sent to the"
                  " log file:\n\t\"%s\"\n\n", global.log_file);
#endif

  remove(global.log_file);
  fd = open(global.log_file, O_RDWR|O_CREAT, 0644);
  if(fd < 0){
    perror("open()");
    return -1;
  }
  close(1);
  if(dup2(fd, 1) < 0){
    perror("dup2()");
    return -1;
  }
  close(2);
  if(dup2(fd, 2) < 0){
    perror("dup2()");
    return -1;
  }
  close(fd);

  return 0;
}

#ifdef NWS
int  initforecasting1()
{ 
  int ok;

  /* forcaster initialisation */ 
  ok=NWSAPI_UseForecaster(NWSAPI_MakeHostSpec(
       global.my_self->host_desc->hostname,NWS_DEFAULT_FORECASTER_PORT));

  if (ok==0){
    fprintf(stderr,"ERROR: NWSUSeForecast ...\n");
    return -1;
  }

  /* which nameserver */
  ok=NWSAPI_UseNameServer(NWSAPI_MakeHostSpec(
      global.my_self->host_desc->hostname, NWS_DEFAULT_NAMESERVER_PORT));

  if (ok==0){
    fprintf(stderr,"ERROR: NWSnameserver ...\n");
    return -1;
  }

  /*which memory */
  ok=NWSAPI_UseMemory(NWSAPI_MakeHostSpec(
          global.my_self->host_desc->hostname, NWS_DEFAULT_MEMORY_PORT));

  if (ok==0){
    fprintf(stderr,"ERROR: NWSmem ...\n");
    return -1;
  }
 
  return 0;
}

int  initforecasting2()
{ 
  int ok;

  /* forcaster initialisation */ 
  ok=NWSAPI_UseForecaster(NWSAPI_MakeHostSpec(
      global.my_self->host_desc->hostname,NWS_DEFAULT_FORECASTER_PORT));

  if (ok==0){
    fprintf(stderr,"ERROR: NWSAPIUSeForecast ...\n");
    return -1;
  }

  /* which nameserver */
  ok=NWSAPI_UseNameServer(NWSAPI_MakeHostSpec(
      global.master->host_desc->hostname, NWS_DEFAULT_NAMESERVER_PORT));

  if (ok==0){
    fprintf(stderr,"ERROR: NWSnameserver ...\n");
    return -1;
  }

  /*which memory */
  ok=NWSAPI_UseMemory(NWSAPI_MakeHostSpec(
          global.master->host_desc->hostname, NWS_DEFAULT_MEMORY_PORT));

  if (ok==0){
    fprintf(stderr,"ERROR: NWSmem ...\n");
    return -1;
  }

  return 0;
}

/* setNWS() */
int setNWS()
{
  int pid;

  fprintf(stderr, "Starting NWS ...\n");

  if (global.master==NULL) /* I am the master agent, start nameserver, */
                              /* a memory and a forecaster*/
  {
    fprintf(stderr, "Starting NWS name server ...\n");

    pid=startNWSNameServer(global.my_self->host_desc->hostname);
    if ( pid < 0){
      fprintf(stderr, "Error starting nameserver\n");
      return -1;
    }

    fprintf(stderr, "Starting NWS memory ...\n");
    pid=startNWSMemory(
         global.my_self->host_desc->hostname,NWS_DEFAULT_NAMESERVER_PORT,
         global.my_self->host_desc->hostname);

    if (pid < 0){
      fprintf(stderr, "Error starting memory\n");
      return -1;
    }

    fprintf(stderr, "Starting NWS forecaster ...\n");
    pid=startNWSForecaster(global.my_self->host_desc->hostname,
                           NWS_DEFAULT_NAMESERVER_PORT);

    if (pid < 0) {
      fprintf(stderr, "Error starting forecaster\n");
      return -1;
    }

    if(initforecasting1()==-1){
      fprintf(stderr, "Error initializing forecast\n");
      return -1;
    }

    return 0;
  }

  /* I start only my own forecaster */
  fprintf(stderr, "Starting NWS forecaster ...\n");
  pid=startNWSForecaster(global.master->host_desc->hostname,
                         NWS_DEFAULT_NAMESERVER_PORT);

  if (pid < 0){ 
    fprintf(stderr, "Error starting forecaster\n");
    return -1;
  }

  if(initforecasting2()==-1){
    fprintf(stderr, "Error initializing forecast\n");
    return -1;
  }

  fprintf(stderr," NWS (forecaster) running ...\n");
  return 0;
}
#endif
