/*****************************************************************/
/*      agent_jobid.c                                            */
/*      Jakob Oestergaard                                        */
/*---------------------------------------------------------------*/
/*  Basic jobid management routines                              */
/*****************************************************************/

#include "agentglobal.h"
#include <stdlib.h>
#include <stdio.h>
#include <time.h>

/*
 * A job_id is unique in the NetSolve system
 * (meaning the agent(s), servers, clients, etc.)
 * at any given point in time.  
 *
 * However, they will eventually get recycled.
 * Curretly this means that we will recycle
 * the job_id after every 2^32-1 job submissions.
 * Also, this will limit the NetSolve system to
 * only be able to handle 2^32-1 concurretly 
 * running jobs at any time - however this limitation
 * is fairly academic I guess...  ;-)
 *
 * The only "magic" part is that job_id 0 is
 * reserved and will never be returned by
 * the jobid_alloc() routine.
 */

typedef struct _NS_JobID_Instance {
  int job_id;
  time_t t_start;
  struct _NS_JobID_Instance * next;
} NS_JobID_Instance;

static NS_JobID_Instance * jobid_list = 0;
static int last_job_id = 10000;

/*
 * Just allocate the next available unused
 * job_id.
 */
int jobid_alloc(void)
{
  NS_JobID_Instance * iter;
  /*
   * Find a good ID
   * (yes this breaks if there are 2^32-1 jobs
   *  active, but...)
   * XXX: Maybe there could be DoS or security concerns
   * here, by allocating the numbers consecutively...
   * Maybe it would be better to use something along the
   * lines of the URID in the InfoServer, where the numbers
   * are pseudo-random (but still unique).
   */
  do {
    last_job_id ++;
    if(!last_job_id) last_job_id++;
    for(iter = jobid_list; iter; iter = iter->next)
      if(iter->job_id == last_job_id) break;
  } while(iter);
  /*
   * Remember that we allocated this ID
   */
  if(!(iter = (NS_JobID_Instance*)malloc(sizeof(NS_JobID_Instance)))) {
    fprintf(stderr, "Out of memory in jobid_alloc()\n");
    return 0;
  }
  iter->job_id = last_job_id;
  iter->t_start = time(0);
  iter->next = jobid_list;
  jobid_list = iter;
  return last_job_id;
}

int jobid_exist(int job_id)
{
  NS_JobID_Instance * iter;
  for(iter = jobid_list; iter; iter = iter->next)
    if(iter->job_id == job_id) return 1;
  return 0;
}

void jobid_remove(int job_id)
{
  NS_JobID_Instance * iter, * prev = 0;
  /* Find element */
  for(iter = jobid_list; iter; iter = iter->next)
    if(iter->job_id == job_id) break;
    else prev = iter;
  /* Free it if it exists */
  if(!iter) return;
  if(prev)
    prev->next = iter->next;
  else
    jobid_list = iter->next;
  free(iter);
}

/*
 * Go thru our list of jobs and see
 * which ones are too old to reasonably
 * still exist in the system
 */
void jobid_tidy(void)
{
  time_t now = time(0);
  NS_JobID_Instance * iter, * prev = 0;
  unsigned n_cleans = 0;
  /*
   * Just reap what is too old
   */
  iter = jobid_list;
  while(iter) {
    NS_JobID_Instance * next = iter->next;
    if(now - iter->t_start > JOBID_TTL) {
      if(prev)
	prev->next = next;
      else
	jobid_list = next;
      free(iter);
      n_cleans ++;
    }
    iter = next;
  }
  if(n_cleans)
    fprintf(stderr, "jobid_tidy(): Cleaned %u old jobs\n", n_cleans);
}


