/*****************************************************************/
/*      netslX_assignment.c                                      */
/*      Version 1.0                      Henri Casanova          */
/*---------------------------------------------------------------*/
/*  netslX_assignment()                                          */
/*****************************************************************/

/*
 * netslX_assignment() : vararg C Interface to the NetSolve Client
 */

#include <stdarg.h>
#include "core.h"
#include "client.h"
#include "mnemonics.h"
#include "callingsequence.h"
#include "cfortranclient.h"

int netslX_assignment(NS_IPaddr_type IPaddr,NS_ProblemDesc *pd,NS_VoidStarOrInt *call)
{
  int i;
  int status;
  void **tmp_output_ptr;
  int elapsed;

  /* Initializing the major    */
  setMajorDefault("Row");

  /* Initializing the objects */
  for (i=0;i<pd->nb_input_objects;i++)
  {
    int object_type,data_type;
    object_type = pd->input_objects[i]->object_type;
    data_type = pd->input_objects[i]->data_type;
    freeObject(pd->input_objects[i]);
    pd->input_objects[i] = createEmptyObject(object_type,data_type);
  }
  for (i=0;i<pd->nb_output_objects;i++)
  {
    int object_type,data_type;
    object_type = pd->output_objects[i]->object_type;
    data_type = pd->output_objects[i]->data_type;
    freeObject(pd->output_objects[i]);
    pd->output_objects[i] = createEmptyObject(object_type,data_type);
  }

  /* Processing the constants */

  status = assignConstantValues(pd,pd->input_objects,pd->output_objects);
  if (status == -1)
  {
    freeProblemDesc(pd);
    return ns_errno;
  }

  /* Parsing the stack */
  if (arrayToObjects(CALL_FROM_C,pd,pd->input_objects,pd->output_objects,call) == -1)
  {
    freeProblemDesc(pd);
    return ns_errno;
  }

  /* Processing the formulae */
  status = assignFormulaValues(pd,pd->input_objects,pd->output_objects);
  if (status == -1)
  {
    freeProblemDesc(pd);
    return ns_errno;
  }

  /* Special preprocessing for the packed file type */
  for (i=0;i<pd->nb_input_objects;i++)
  {
    if (pd->input_objects[i]->object_type != NETSOLVE_PACKEDFILES)
      continue;
    pd->input_objects[i]->attributes.packedfiles_attributes.filenames =
         copyCharStarList(
            pd->input_objects[i]->attributes.packedfiles_attributes.filenames,
            pd->input_objects[i]->attributes.packedfiles_attributes.m);
  }


  for (i=0;i<pd->nb_output_objects;i++)
  {
    if (pd->output_objects[i]->object_type == NETSOLVE_PACKEDFILES)
    {
      pd->output_objects[i]->attributes.packedfiles_attributes.defaultprefix =
        (char *)(pd->output_objects[i]->attributes.packedfiles_attributes.filenames);
      pd->output_objects[i]->attributes.packedfiles_attributes.filenames = NULL;
    }
  }

  /* Setting the major */
  for (i=0;i<pd->nb_input_objects;i++)
  {
    if (pd->input_objects[i]->object_type == NETSOLVE_MATRIX)
      pd->input_objects[i]->attributes.matrix_attributes.major = my_major;
    else if(pd->input_objects[i]->object_type == NETSOLVE_SPARSEMATRIX)
      pd->input_objects[i]->attributes.sparsematrix_attributes.major = my_major;
    else
      continue;
  }
 
  for (i=0;i<pd->nb_output_objects;i++)
  {
    if (pd->output_objects[i]->object_type == NETSOLVE_MATRIX)
      pd->output_objects[i]->attributes.matrix_attributes.major = my_major;
    else if(pd->output_objects[i]->object_type == NETSOLVE_SPARSEMATRIX)
      pd->output_objects[i]->attributes.sparsematrix_attributes.major = my_major;
    else
      continue;
  }

  /* Saving the tmp_output_ptr */
  tmp_output_ptr = (void **)calloc(pd->nb_output_objects,sizeof(void*));
  for (i=0;i<pd->nb_output_objects;i++)
  {
    if (pd->output_objects[i]->object_type == NETSOLVE_STRING)
      tmp_output_ptr[i] = 
        pd->output_objects[i]->attributes.string_attributes.ptr;
  }

  /* Calling the low-level blocking solve */

  status = netsolveBlockingAssignment(IPaddr, pd,pd->input_objects,
                       pd->output_objects,&elapsed);

  if (status != -1)
  {
    /* take care of the strings in output */
    for (i=0;i<pd->nb_output_objects;i++)
    {
      if (pd->output_objects[i]->object_type != NETSOLVE_STRING)
        continue;
      strcpy(tmp_output_ptr[i],
          pd->output_objects[i]->attributes.string_attributes.ptr);
      free(pd->output_objects[i]->attributes.string_attributes.ptr);
    }
    ns_errno = NetSolveOK;
  }
  free(tmp_output_ptr);
  return ns_errno;
}
