/*****************************************************************/
/*      waitrequest.c                                            */
/*      Henri Casanova                                           */
/*      Dieter Bachmann                                          */
/*---------------------------------------------------------------*/
/*****************************************************************/

#include "core.h"
#include "client.h"
#include "client2proxy.h"
#include "client2server.h"
#include "netsolve.h"
#include <unistd.h>
#include <signal.h>

/*
 * netsolveWaitRequest()
 * 
 * This function waits for the completion of a request and automatically
 * resubmit the job in case of an error.
 */
int netsolveWaitRequest_sequence(NS_Node *top_node, NS_RequestDesc **rd,int *elapsed)
{
  int status;
  NS_RequestDesc *newrd;
  NS_Node * counter;

  /*
   * When we await a sequence completion, we will receive a number
   * of PartialCompletionOK messages from the proxy (which in
   * turn received them from the server).
   * For each of these we will tell the agent about the partial
   * completion (a partial completion is the completion of a
   * single problem in a sequence of problems).
   */
  counter = top_node;
  do {
    status = CP_awaitJobCompletion(*rd);
    fprintf(stderr, "%i, job completed with status = %i, ns_errno = %i, job_id = %i\n", 
	    getpid(), status, ns_errno, (*rd)->pd->job_id);
    /*
     * In case of a normal partial completion, notify the
     * agent
     */
    if(status == 0 && ns_errno == NetSolvePartialCompletionOK) {

      CP_reportJobCompleted((*rd)->IPaddr,(*rd)->request_id,(*rd)->pd->job_id);

    }

    /*
     * We don't care whether counter points to the actual
     * node completed... All we want is to know when we have
     * received N completions from the N node sequence, therefore
     * it is sufficient to simply traverse thru the sequence
     * waiting for *some* completion for each node. When we're
     * done with the nodes, we're done with the sequence.
     */
    if(!(counter = counter->next))
      break;

  } while(status == 0 && ns_errno == NetSolvePartialCompletionOK);

  fprintf(stderr, "%i, WE QUIT WAITREQUEST: status = %i, ns_errno = %i\n",
	  getpid(), status, ns_errno);

  /*
   * Let the proxy know that this ID (proxy-unique ID) is no longer
   * in use
   */
  CP_retireJob(*rd);

  /*
   * Check job completion status
   */
  if (status == -1) {
    /* Resubmit */
    if (ns_errno == NetSolveInvalidRequestID) {
      *rd = NULL;
      *elapsed = 0;
      return -1;
    }
    reportFailure((*rd)->IPaddr, (*rd)->pd->nickname,
		  (*rd)->request_id,SERVER_ERROR,(*rd)->pd->job_id);
    newrd = netsolve_sequence_SendRequest(top_node);
    free(*rd);
    *rd = newrd;
    if (*rd == NULL)
      return -1;
    return netsolveWaitRequest(rd,elapsed);
  }
   
  if (CS_downloadOutput_sequence(top_node, *rd,elapsed) == -1)
  { 
    switch (ns_errno) 
    {
      case NetSolveBadValues:
        return -1;
      case NetSolveDimensionMismatch:
        return -1;
      case NetSolveNoSolution:
        return -1;
    }
    /* Resubmit */
    reportFailure((*rd)->IPaddr,(*rd)->pd->nickname,
                  (*rd)->request_id,HOST_ERROR,(*rd)->pd->job_id);
    newrd = netsolve_sequence_SendRequest(top_node);
    free(*rd);
    *rd = newrd;
    if (*rd == NULL)
      return -1;
    return netsolveWaitRequest_sequence(top_node, rd, elapsed); 
  }

  CP_reportJobCompleted((*rd)->IPaddr,(*rd)->request_id,(*rd)->pd->job_id);

  ns_errno = NetSolveOK;
  return 1; 
} /* int netsolveWaitRequest_sequence(NS_Node *top_node, RequestDesc **rd,int *elapsed) */


