/*****************************************************************/
/*   Multinode sequencing host selection for the proxy           */
/*                                                               */
/*    Jakob Oestergaard                                          */
/*****************************************************************/

#include "node.h"
#include "netsolveproxy.h"
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

#define MULTINODE_DEBUG 1

/*
 * Declarations and utility stuff
 */
typedef struct _GroupNode {
  NS_IPaddr_type hostaddr;  /* key field */
  NS_IPaddr_type ibpaddr;   /* value field */
  struct _GroupNode * next;
} GroupNode;

GroupNode * new_GroupNode(const NS_IPaddr_type hostaddr, 
			  const NS_IPaddr_type ibpaddr)
{
  GroupNode * n = (GroupNode*)malloc(sizeof(GroupNode));
  if(!n) {
    fprintf(stderr, "Out of memory in GroupNode allocation!\n");
    return 0;
  }
  n->hostaddr = hostaddr;
  n->ibpaddr = ibpaddr;
  n->next = 0;
  return n;
}

void delete_GroupNode(GroupNode * n)
{
  n->next = 0;
  free(n);
}

void delete_set_GroupNode(GroupNode * n)
{
  GroupNode * next;
  while(n) {
    next = n->next;
    delete_GroupNode(n);
    n = next;
  }
}

/*
 * We implement a set with a sorted linked list because we're lazy
 * and cycles are cheap.
 */
GroupNode * insert_GroupNode(GroupNode * orig, GroupNode * n)
{
  /* Insertion in an empty set yields the inserted set */
  if(!orig) {
    return n;
  }
  while(n) {
    GroupNode * nextn = n->next;
    GroupNode * iterator = orig;
    GroupNode * piter = 0;
    /* find insertion point */
    while(iterator && n->hostaddr > iterator->hostaddr) {
      piter = iterator;
      iterator = iterator->next;
    }
    /* don't insert duplicate nodes */
    if(iterator && n->hostaddr == iterator->hostaddr) {
      delete_GroupNode(n);
      return orig;
    }
    /* insertion */
    if(piter) {
      /* chain in */
      piter->next = n;
      n->next = iterator;
    } else {
      /* insert in front */
      orig = n;
      n->next = iterator;
    }
    /* next */
    n = nextn;
  }
  return orig;
}

/*
 * Return the groupnode holding the host IP given,
 * or return 0.
 */
const GroupNode * find_GroupNode(const GroupNode * group, NS_IPaddr_type host)
{
  const GroupNode * iterator;
  for(iterator = group; iterator; iterator = iterator->next)
    if(iterator->hostaddr == host)
      break;
  return iterator;
}

/*
 * Count number of elements in group
 */
unsigned size_GroupNode(const GroupNode * group)
{
  unsigned count = 0;
  const GroupNode * iterator;
  for(iterator = group; iterator; iterator = iterator->next)
    count++;
  return count;
}


#ifdef MULTINODE_DEBUG
void print_group_members(GroupNode** groups, unsigned ngroups, NS_Node * top_node)
{
  GroupNode * giter;
  unsigned gi;
  NS_Node * iterator;

  for(gi = 0; gi != ngroups; gi++) {
    fprintf(stderr, "*> -------  group %u members -------\n", gi);
    for(giter = groups[gi]; giter; giter = giter->next) {
      fprintf(stderr, "*>  server(%lu) <-> ibp(%lu) : {",
	      (unsigned long)giter->hostaddr, (unsigned long)giter->ibpaddr);
      for(iterator = top_node; iterator; iterator = iterator->next) {
	unsigned ai;
	for(ai = 0; ai < iterator->servers->nb_servers; ai++) {
	  if(giter->hostaddr == iterator->servers->IPaddrlist[ai])
	    fprintf(stderr, " %u", iterator->pd->job_id);
	}
      }
      fprintf(stderr, " }\n");
    }
    fprintf(stderr, "*> ---------------------------------\n");
  }
}
#endif

/*
 * Given the list of nodes in a sequence, we will select a list of
 * machines (also called nodes) that can be used for execution of
 * each (sequence-) node.
 * The idea is, that the master-server given responsibility for the sequence
 * (which must be one of the servers in the top_node->servers list) can
 * bring in slave-servers for execution of single nodes in the sequence as
 * it sees fit.
 * This routine should look at the servers suggested for each node, and
 * decide on a set of servers that both covers all problems (nodes) and
 * share a local IBP server.
 */
NS_Node * select_multinode_nodes(NS_Node * top_node)
{
  NS_Node * iterator;
  GroupNode ** groups;
  unsigned ngroups;
  unsigned gi;
  unsigned maxgroup_count, maxgroup_index;
  
  /* There can never be more groups than there are servers in the top node */
  groups = (GroupNode**)malloc(top_node->servers->nb_servers * sizeof(GroupNode*));
  if(!groups) {
    fprintf(stderr, "Out of memory allocating group array for %i groups\n", top_node->servers->nb_servers);
    return 0;
  }

  /* Now build up the groups...
   *
   * Group up servers so we have groups that
   * can handle all nodes in the sequence, but where
   * all servers in one group share the same local
   * IBP server.
   *
   * For each node:
   *  For each server:
   *   place in group corrosponding to IBP server.
   *
   */
  ngroups = 0;
  for(iterator = top_node; iterator; iterator = iterator->next) {
    unsigned si;
    for(si = 0; si < iterator->servers->nb_servers; si++) {
      /*
       * Find group... (groups are guaranteed to be non-empty)
       */
      GroupNode * gn;
      unsigned gi;
      for(gi = 0; gi != ngroups; gi++)
	if(groups[gi]->ibpaddr == iterator->servers->IBPaddrlist[si])
	  break;
      /*
       * If  group couldn't be found, create new empty one
       */
      if(gi == ngroups) {
	groups[ngroups++] = 0;
      }
      /*
       * Insert server in group
       */
      gn = new_GroupNode(iterator->servers->IPaddrlist[si],
			 iterator->servers->IBPaddrlist[si]);
      if(!gn) {
	fprintf(stderr, "Skipping new group %u beause of memory shortage...\n", gi);
	ngroups--;
	continue;
      }
      groups[gi] = insert_GroupNode(groups[gi], gn);
    }
  }

  /*
   * Now that we have the IBP groups, we should weed out the ones
   * that are unfit for the sequence...
   *
   * For each group:
   *  For each node:
   *   If the group does not contain at least one of the servers in the
   *   - server-list of the node, kill the group (it is unfit)
   *
   */
  for(gi = 0; gi != ngroups; gi++) {
    for(iterator = top_node; iterator; iterator = iterator->next) {
      unsigned hosti;
      /*
       * See if at least on of the capable servers are found in this
       * group.
       */
      for(hosti = 0; hosti < iterator->servers->nb_servers; hosti++)
	if(find_GroupNode(groups[gi], iterator->servers->IPaddrlist[hosti]))
	  break;
      /*
       * If not, kill the group.
       */
      if(hosti == iterator->servers->nb_servers) {
	delete_set_GroupNode(groups[gi]);
	groups[gi] = 0;
	continue;
      }
    }
  }

#ifdef MULTINODE_DEBUG
  fprintf(stderr, "*> Total groups created\n");
  print_group_members(groups, ngroups, top_node);
#endif

  /*
   * Finally, now we have either zero or more groups left (surprise!), if
   * we have zero groups, we should see if we can execute the sequence on
   * a set of non-local servers...  XXX : do this later.
   * We assume we have >=1 groups left for now...  Then we should select
   * the best group.
   *
   * For each group we calculate the CPU power available...  Note, the
   * groups are no longer guaranteed to be non-empty !
   */

  maxgroup_count = 0;
  maxgroup_index = 0;
  for(gi = 0; gi != ngroups; gi++) {
    unsigned cnt;
    /*
     * XXX:  We should send the list of groups to the agent and
     * let the agent assign a group for us to use...
     * For now, we just use the biggest group
     */
    if((cnt = size_GroupNode(groups[gi])) > maxgroup_count) {
      maxgroup_count = cnt;
      maxgroup_index = gi;
    }
  }
  gi = maxgroup_index;

  if(!maxgroup_count) {
    fprintf(stderr, "Sorry, no IBP groups available for sequence !!\n");
    return 0;
  }

  /*
   * Now step thru the sequence and assign servers for each node...
   */  
  for(iterator = top_node; iterator; iterator = iterator->next) {
    /*
     * For each server:  If it does not exist in the chosen group,
     * remove it from the list of possible servers
     */
    unsigned si;
    for(si = 0; si < iterator->servers->nb_servers;) {
      if(!find_GroupNode(groups[gi], iterator->servers->IPaddrlist[si])) {
	/* Ok, remove this server from the list in this node */
	unsigned addri;
	fprintf(stderr, "For job_id %u we kill server %lu\n", iterator->pd->job_id,
		(unsigned long)iterator->servers->IPaddrlist[si]);
	/* We remove the server from the array of addresses, thus this loop... */
	for(addri = si; addri < iterator->servers->nb_servers-1; addri++) {
	  free(iterator->servers->hostlist[addri]);
	  iterator->servers->hostlist[addri] = iterator->servers->hostlist[addri+1];
	  iterator->servers->IPaddrlist[addri] = iterator->servers->IPaddrlist[addri+1];
	  iterator->servers->portlist[addri] = iterator->servers->portlist[addri+1];
	  iterator->servers->dataformatlist[addri] = iterator->servers->dataformatlist[addri+1];
	  iterator->servers->predictedtimelist[addri] = iterator->servers->predictedtimelist[addri+1];
	  iterator->servers->IBPaddrlist[addri] = iterator->servers->IBPaddrlist[addri+1];
	}
	iterator->servers->nb_servers--;
      } else {
	/* We only increment the counter if we didn't remove a server */
	si++;
      }
    }
  }

  /*
   * Now each node in the sequence holds the list of servers on which it
   * can execute, and we know that all servers for all nodes share the
   * same local IBP server.  So far, so good (and so what?)
   */

#ifdef MULTINODE_DEBUG
  fprintf(stderr, "*> Selected group %i for sequence\n", gi);
  print_group_members(groups, ngroups, top_node);
#endif

  /*
   * Clean up
   */
  for(gi = 0; gi != ngroups; gi++)
    delete_set_GroupNode(groups[gi]);
  
  return top_node;
}

