/*****************************************************************/
/*      nsprox_process_messages.c                                */
/*      Henri Casanova                                           */
/*---------------------------------------------------------------*/
/*****************************************************************/

#include "core.h"
#include "client.h"
#include "proxyjobdesc.h"
#include "proxy.h"
#include "netsolvesignals.h"
#include "agentprocessmessage.h"
#include "netsolveerror.h"
#include <sys/types.h>
#include <netinet/in.h>
#include <unistd.h>

/*
 * proxyProcessRequest()
 *
 * This function implements all the system-independent proxy 
 * functionalities and calls the system-specific functions. Those
 * functions are implemented in sub-directories (e.g. NetSolve, Globus)
 */
void proxyProcessRequest(NS_Communicator *comm,int listeningsocket)
{
  int rc;
  int tag;

  if (recvInt(comm,&tag) == -1)
  {
    netsolvePerror(__FILE__,__LINE__);
    return;
  }
  
  /* No fork: some of the proxy functions opperate on the 
   * state of the proxy and must therefore no be performed
   * by another process
   */
  if (tag == NS_PROT_DIE)
  {
    close(listeningsocket);
    endTransaction(comm);
    exit(0);
  } 
  if (tag == NS_PROT_NEW_JOB_DESC)
  {
    processNewJobDesc(comm);
    endTransaction(comm);
    return;
  }
  if (tag == NS_PROT_RM_JOB_DESC)
  {
    processRmJobDesc(comm);
    endTransaction(comm);
    return;
  }
  if (tag == NS_PROT_JOB_REPORT)
  {
    processJobReport(comm);
    endTransaction(comm);
    return;
  }
  if (tag == NS_PROT_JUST_CHECKING) {
    endTransaction(comm);
    return;
  }
  if (tag == NS_PROT_AWAIT_COMPLETION) {
    processAwaitCompletion(comm);
    /* important: do *NOT* end the transaction for this particular
     * call because the communicator will be used and freed later on
     */
    return;
  }

  /* fork for everything else */
  if (-1 == (rc = fork())) {
    /* fork failed */
    /* XXX: We should not give a protocol error -
     * something else should be introduced for resource
     * related errors.
     */
    if (sendInt(comm,NS_PROT_PROTOCOL_ERROR) == -1)
      netsolvePerror(__FILE__,__LINE__);
    endTransaction(comm);
    return;
  }
  /* parent returns imediately */
  if(rc) {
    endTransaction(comm);
    return;
  }
 
  close(listeningsocket);

  switch(tag)
  {
    case NS_PROT_NB_SERVERS:
      processNumberOfServers(comm);
      break;
    case NS_PROT_AGENT_LIST:
      processAgentList(comm);
      break;
    case NS_PROT_SERVER_LIST:
      processServerList(comm);
      break;
    case NS_PROT_PROBLEM_LIST:
      processProblemList(comm);
      break;
    case NS_PROT_PROBLEM_INFO:
      processProblemInfo(comm);
      break;
    case NS_PROT_SV_FAILURE:
      processServerFailure(comm);
      break;
    case NS_PROT_KILL_JOB:
      processKillJob(comm);
      break;
    case NS_PROT_SEND_ASSIGNMENT:
      processSendAssignment(comm);
      break;
    case NS_PROT_SEND_REQUEST:
      processSendRequest(comm);
      break;
    case NS_PROT_SEND_SEQUENCE_REQUEST:
      processSend_sequence_Request(comm);
      break;
    case NS_PROT_PROBE_REQUEST:
      processProbeRequest(comm);
      break;
    case NS_PROT_JOB_COMPLETED:
      processJobCompleted(comm);
      break;
  case NS_PROT_ALLOC_JOBID:
    processAllocJobId(comm);
    break;
    default: /* ERROR */
      if (sendInt(comm,NS_PROT_PROTOCOL_ERROR) == -1)
        netsolvePerror(__FILE__,__LINE__);
      break;
  }
  endTransaction(comm); 
  exit(0);
}

/*
 * processJobReport()
 *
 * One of my children has been contacted by a server that reported
 * its status. This child then reports it to me so that it is possible
 * to update the global state.
 */
void processJobReport(NS_Communicator *comm)
{
  int job_id;
  int ID;
  int error_code;
  int foo;
  NS_NotificationWaiter * iter, * piter = 0;

  if ((recvInt(comm,&job_id) == -1) ||    /* checked - ok - matches Server/service_util.c:27 */
      (recvInt(comm,&ID) == -1) ||
      (recvInt(comm,&error_code) == -1))
  { 
    netsolvePerror(__FILE__,__LINE__);
    return;
  } 
  if ((ID<0)||(ID>NB_MAX_JOBS)) {
    fprintf(stderr, "Skipping job report with ID=%i because it's out of range\n", ID);
    return;
  }
  if (jobs[ID] == NULL) {
    fprintf(stderr, "Skipping job report with ID=%i because job is nonexistent\n", ID);
    return;
  }
 
  jobs[ID]->job_id = job_id;
  jobs[ID]->error_code = error_code;

  fprintf(stderr, "Got job report with error_code=%i and ID=%i and job_id=%i\n",
	  error_code, ID, job_id);

  /*
   * If this was not a job completion report, don't look tru
   * the waiter list and send back a reply to whoever
   * cares about this
   */
  if(error_code == NetSolveNotReady)
    return;

  for(iter = notification_waiters; iter; iter = iter->next) {
    if(iter->id == ID) {
      /* We found it !
       * Send back result
       */
      fprintf(stderr, "Sending completion event to fd = %i\n",
	      iter->comm->sock);
      if ((sendInt(iter->comm, NetSolveOK) == -1) ||
	  (sendInt(iter->comm, WAITMAGIC) == -1) ||
	  (sendInt(iter->comm, error_code) == -1) ||
	  (sendInt(iter->comm, job_id) == -1)) {
	netsolvePerror(__FILE__, __LINE__);
      } else {
	/*
	 * Get ack
	 */
	recvInt(iter->comm, &foo);
	/* if this was a NetSolvePartialCompletionOK, then don't
	 * unchain  (we're processing a sequene here...)
	 */
	if(error_code == NetSolvePartialCompletionOK) {
	  fprintf(stderr, "In processJobReport(): Got partial completion so resetting error_code\n");
	  /*
	   * Since this was a partial completion only, we should set
	   * back the error code to unfinished...
	   */
	  jobs[ID]->error_code = NetSolveNotReady;
	  break;
	}
      }
    }
    piter = iter;
  }
}

/*
 *  processNewJobDesc()
 *
 *  One of my childs has successfully launched a job and 
 *  asks me for a new job descriptor ID.
 */
void processNewJobDesc(NS_Communicator *comm)
{
  NS_JobDesc *jd;
  int ID;

  if ((jd = recvJobDesc(comm)) == NULL)
  {
    netsolvePerror(__FILE__,__LINE__);
    return;
  }
  
  ID = insertJobDesc(jd);

  if (sendInt(comm,ID) == -1)
  {
    netsolvePerror(__FILE__,__LINE__);
    return;
  }
  return; 
}

/*
 * processRmJobDesc()
 *
 * one of my child has been asked to remove a job descriptor.
 */
void processRmJobDesc(NS_Communicator *comm)
{
  int ID;
  NS_NotificationWaiter * iter, * piter;

  if (recvInt(comm,&ID) == -1)
  {
    netsolvePerror(__FILE__,__LINE__);
    return;
  }
  if ((ID<0)||(ID>NB_MAX_JOBS))
    return;
  if (jobs[ID] == NULL)
    return;
  freeJobDesc(jobs[ID]);
  fprintf(stderr, "processRmJobDesc(): Killing job ID %i\n", ID);
  jobs[ID] = NULL;

  /*
   * Go thru the waiter list and see if we have stuff there to clean as
   * well...
   */
  piter = 0;
  for(iter = notification_waiters; iter; iter = iter->next) {
    if(iter->id == ID) {
      /* kill the comms channel */
      endTransaction(iter->comm);
      /* Now unchain and free */
      if(!piter)
	notification_waiters = iter->next;
      else
	piter->next = iter->next;
      free(iter);
      break;
    } else {
      piter = iter;
    }
  }

  return;
}

/*
 * We should ask the agent to allocate
 * a job_id for the client problem...
 * (if we did direct client<->agent
 *  comms. this routine would not be needed)
 */
void processAllocJobId(NS_Communicator * comm_to_client)
{
  NS_Communicator * comm_to_agent;
  int result;

  /* contact the agent */
  if(!(comm_to_agent = contactHost(agent,0,AGENT_PORT,DATA_XDR))) {
    sendInt(comm_to_client,NetSolveCannotContactAgent);
    netsolvePerror(__FILE__,__LINE__);
    return;
  }
   
  /* relay request to agent */
  if(sendInt(comm_to_agent,NS_PROT_ALLOC_JOBID) == -1) {
    sendInt(comm_to_client,NetSolveNetworkError);
    netsolvePerror(__FILE__,__LINE__);
    endTransaction(comm_to_agent);
    return;
  }
  
  /* get reply from agent */
  if(recvInt(comm_to_agent,&result) == -1) {
    sendInt(comm_to_client,NetSolveNetworkError);
    netsolvePerror(__FILE__,__LINE__);
    endTransaction(comm_to_agent);
    return;
  }
  if(result != NetSolveOK) {
    sendInt(comm_to_client,NetSolveInternalError);
    netsolvePerror(__FILE__,__LINE__);
    endTransaction(comm_to_agent);
    return;
  }

  if(recvInt(comm_to_agent,&result) == -1) {
    sendInt(comm_to_client,NetSolveNetworkError);
    netsolvePerror(__FILE__,__LINE__);
    endTransaction(comm_to_agent);
    return;
  }
  if(result == 0) {
    sendInt(comm_to_client,NetSolveInternalError);
    netsolvePerror(__FILE__,__LINE__);
    endTransaction(comm_to_agent);
    return;
  }

  endTransaction(comm_to_agent);

  /* relay reply to client */
  if(-1 == sendInt(comm_to_client,NetSolveOK)) {
    netsolvePerror(__FILE__,__LINE__);
    return;
  }
  if(-1 == sendInt(comm_to_client,result)) {
    netsolvePerror(__FILE__,__LINE__);
    return;
  }

  fprintf(stderr, "Proxy relayed a newly allocated jobid=%d (0x%X) to client\n",
	  (unsigned)result, (unsigned)result);

  /* we're done */
}

