/*****************************************************************/
/*      petscutil.c                                              */
/*      Henri Casanova                                           */
/*****************************************************************/

#include "core.h"
#include "petscservice.h"
#include <fcntl.h>
#include <math.h>

/*
 * newPETSCVectorDesc()
 */
int *newPETSCVectorDesc(int first,int last)
{
  int *new;
  new = (int *)calloc(2,sizeof(int));
  new[0] = first;
  new[1] = last;
  return new;
}

/*
 * petsc_exit()
 */
void petsc_exit(char *s)
{
  if (s == NULL)
    createDoneFile("ERROR");
  else
    createDoneFile(s);
  MPI_Finalize();
  exit(0);
}

/*
 * PETSCReadInputObjectsFromFiles()
 *
 * Reads the input from file. Only the Matrices are 2-D
 * block distributed fo now.
 */
int PETSCReadInputObjectsFromFiles(NS_ProblemDesc *pd)
{
  int i;
  char buffer[256];
  int fd;
  NS_Object *obj;

  for (i=0;i<pd->nb_input_objects;i++)
  {
    sprintf(buffer,"./input%d",i);
    obj = pd->input_objects[i];
    switch(obj->object_type)
    {
      case NETSOLVE_VECTOR:
        if (read1DDistributedVector(buffer,obj) == -1)
          return -1;
        break;
      default: /* Every processor need to get the object */
        fd = open(buffer,O_RDONLY,0666);
        if (fd < 0)
        {
#ifdef VIEW
          fprintf(stderr,"Impossible to open file '%s'\n",buffer);
#endif
          return -1;
        }
        if (readObjectFromFile(fd,pd->input_objects[i]) == -1)
        {
          netsolvePerror(__FILE__,__LINE__);
          close(fd);
          return -1;
        }
        close(fd);
        break;
    }
  }
  return 1;
}

/*
 * read1DDistributedVector()
 */
int read1DDistributedVector(char *filename,NS_Object *obj)
{
  int m,local_vector_size;
  void *ptr;
  int fd;
  int dummy;
  int first,last;

  fd = open(filename,O_RDONLY,0666);
  if (fd < 0)
  {
#ifdef VIEW
    fprintf(stderr,"impossible to open file '%s'\n",filename);
#endif
    return -1;
  }

  /* read the object type and the data type */
  if (read(fd,&dummy,sizeof(int)) != sizeof(int))
  {
    close(fd);
    return -1;
  }
  if (read(fd,&dummy,sizeof(int)) != sizeof(int))
  {
    close(fd);
    return -1;
  }
  
  if (read(fd,&m,sizeof(int)) != sizeof(int))
  {
    close(fd);
    return -1;
  }
  close(fd);

  /* Figure out the space */
  if (MYROW != NUM_PROCS-1)
  {
    first = 
    last = 
    local_vector_size = m / NUM_PROCS;
  }
  else
  {
    first = 
    last = 
    local_vector_size = m / NUM_PROCS+ m % NUM_PROCS;
  }

  fprintf(stderr,"local_vector_size = %d",local_vector_size);
   
  obj->attributes.vector_attributes.m = m;
  obj->attributes.vector_attributes.d = 
           newPETSCVectorDesc(first,last);

  /* Allocate the space */
  local_matrix_size = Cnumroc(m,ROW_BLOCK,MYROW,0,PROC_ROWS) *
                      Cnumroc(n,COL_BLOCK,MYCOL,0,PROC_COLS);
  ptr = calloc(local_matrix_size,netsolve_sizeof(obj->data_type));
  obj->attributes.matrix_attributes.ptr = ptr;

  /* construct a descriptor */
  desc = descInit(m,n,1);
  obj->attributes.matrix_attributes.d = desc;

  /* allocate some workspace */
  workspace = (double *)calloc(desc[MB_],netsolve_sizeof(obj->data_type));

  /* Open the file again and call pdlaread */

  /***
   * KIM: I assume only the 1,1 processor should open the file, right ?
   ***/
  globalfile = fopen(filename,"r");
  if (filename == NULL)
  {
#ifdef VIEW
    fprintf(stderr,"Impossible to open file '%s'\n",filename);
#endif
    ns_errno = NetSolveFileError;
    free(workspace);
    return -1;
  }
#ifdef ADD_
  pdlaread_(ptr,desc,&zero,&zero,workspace);
#endif
#ifdef UPCASE
  PDLAREAD(ptr,desc,&zero,&zero,workspace);
#endif
#ifdef NOCHANGE
  pdlaread(ptr,desc,&zero,&zero,workspace);
#endif
   
  fclose(globalfile);
  free(workspace);
  return 1;
}

/*
 * PETSCWriteOutputObjectsToFiles()
 *
 * Writes the output to file. Only the Matrices are 2-D
 * block distributed fo now.
 */
int PETSCWriteOutputObjectsToFiles(int client_major,
          NS_ProblemDesc *pd)
{
  int i;
  char buffer[256];
  int fd;
  NS_Object *obj;

  for (i=0;i<pd->nb_output_objects;i++)
  {
    sprintf(buffer,"./output%d",i);
    obj = pd->output_objects[i];
    switch(obj->object_type)
    {
      case NETSOLVE_MATRIX:
        if (client_major != pd->major) /* transpose ? */
        {
          Cpdtran(obj->attributes.matrix_attributes.m,
                obj->attributes.matrix_attributes.n,
                1.0,
                obj->attributes.matrix_attributes.ptr,
                1,1,
                obj->attributes.matrix_attributes.d,
                0.0,
                obj->attributes.matrix_attributes.ptr,
                1,1,
                obj->attributes.matrix_attributes.d);
          obj->attributes.matrix_attributes.major = client_major;
        }
        
        if (write2DBlockCyclicMatrix(buffer,obj) == -1)
          return -1;
          break;
        break;
      default: /* Processor 0 writes object */
        if ((MYROW!=0)||(MYCOL!=0))
          break;
        fd = open(buffer,O_WRONLY|O_CREAT,0666);
        if (fd < 0)
        {
#ifdef VIEW
          fprintf(stderr,"Impossible to create file '%s'\n",buffer);
#endif
          return -1;
        }
        if (writeObjectToFile(fd,pd->input_objects[i]) == -1)
        {
          netsolvePerror(__FILE__,__LINE__);
          close(fd);
          return -1;
        }
        close(fd);
        break;
    }
  }
  return 1;
}

/*
 * write1DDistributedVector()
 */
int write1DDistributedVector(char *filename,NS_Object *obj)
{
  int m,n,major,local_matrix_size;
  void *ptr;
  int *desc;
  int fd;
  double *workspace;
  int zero = 0;
  int one = 1;

  major = obj->attributes.matrix_attributes.major;
  m = obj->attributes.matrix_attributes.m;
  n = obj->attributes.matrix_attributes.n;
  desc = obj->attributes.matrix_attributes.d;
  ptr = obj->attributes.matrix_attributes.ptr;

  if ((MYROW==0)&&(MYCOL==0))
  {
    fd = open(filename,O_WRONLY|O_CREAT,0666);
    if (fd < 0)
    {
#ifdef VIEW
      fprintf(stderr,"impossible to open file '%s'\n",filename);
#endif
      ns_errno = NetSolveFileError;
      return -1;
    }
  
    if (write(fd,&(obj->object_type),sizeof(int)) != sizeof(int))
    {
      netsolvePerror(__FILE__,__LINE__);
      close(fd);
      return -1;
    }
    if (write(fd,&(obj->data_type),sizeof(int)) != sizeof(int))
    {
      netsolvePerror(__FILE__,__LINE__);
      close(fd);
      return -1;
    }


    /* Writing the sizes */
    if (write(fd,&major,sizeof(int)) != sizeof(int))
    {
      netsolvePerror(__FILE__,__LINE__);
      close(fd);
      return -1;
    }

    if (write(fd,&m,sizeof(int)) != sizeof(int))
    {
      netsolvePerror(__FILE__,__LINE__);
      close(fd);
      return -1;
    }

    if (write(fd,&n,sizeof(int)) != sizeof(int))
    {
      netsolvePerror(__FILE__,__LINE__);
      close(fd);
      return -1;
    }
    close(fd);
  }

  /***
   * KIM: I assume only the 1,1 processor should open the file, right ?
   ***/
  if ((MYROW==0)&&(MYCOL==0))
    globalfile = fopen(filename,"a");
#if 0
  if (globalfile == NULL)
  {
#ifdef VIEW
    fprintf(stderr,"Impossible to open file '%s'\n",filename);
#endif
    free(workspace);
    return -1;
  }
#endif

  workspace = (double *)calloc(desc[MB_],sizeof(double));

#ifdef ADD_
  pdlawrite_(&m,&n,ptr,&one,&one,desc,&zero,&zero,workspace);
#endif
#ifdef UPCASE
  PDLAWRITE(&m,&n,ptr,&one,&one,desc,&zero,&zero,workspace);
#endif
#ifdef NOCHANGE
  pdlawrite(&m,&n,ptr,&one,&one,desc,&zero,&zero,workspace);
#endif
   
  if ((MYROW==0)&&(MYCOL==0))
    fclose(globalfile);
  free(workspace);
  return 1;
}

