/*****************************************************************/
/*      generateservice_sequence.c                               */
/*      Henri Casanova                                           */
/*      Dieter Bachmann                                          */
/*---------------------------------------------------------------*/
/*****************************************************************/

#include "netsolve.h"
#include "core.h"
#include "serverglobal.h"
#include "generateservice.h"
#include "standardservice.h"
#include "condorservice.h"
#include "petscservice.h"
#include "scalapackservice.h"
#include "netsolveupf.h"
#include "serviceutil.h"
#include "lifelink.h"
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <fcntl.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <strings.h>

#ifdef MULTINODESEQ
# include "ibp_base.h"
# include "ibp_client.h"
#endif


/*
 * We need some list handling for multinode spawns
 */
typedef struct _multinode_spawn_e {
  NS_Node * node;
  NS_Communicator * remote;
  struct _multinode_spawn_e * next;
} multinode_spawn_e;

multinode_spawn_e * mspawns = 0;


/*
 * Uncomment the following definition for DAG debugging
 * #define SEQUENCING_DEBUG
 */

#define SEQUENCING_DEBUG 1
#define VIEW 1

static unsigned int running_jobs = 0;

#ifdef MULTINODESEQ
static struct ibp_depot depot = { "", 0 };
#endif

/*
 * generateServiceProcess_sequence
 *
 */
void generateServiceProcess_sequence(NS_Communicator *comm, int restriction_index,
				     NS_Node *top_node,char *agent_name,
				     NS_IPaddr_type proxy_IPaddr, int proxy_port, char * localibp_host)
{

  NS_Node *tmp_node;
  int ID;
  int fd;
  int upf,customized,ftp;
  char tmpdir[256];
  int tag;
  char problemdescfile[256];
  int count=0;
  NS_Socket_type listening_socket;
  NS_Socket_type sock;
  int my_port;
  NS_Communicator *comm_to_client;
  int client_major;
  int my_pid;
  int lifelink_pid;
  int wait_pid;
  int wait_status;
  time_t date1,date2;
  int elapsed;

  /*
   * (jakob) XXX:
   * If fork() fails, we return as if everything
   * was perfect.... This is obviously a mistake.
   */
   if (fork())
       return;  
 
  close(global.sock);

#ifdef MULTINODESEQ
  depot.port = 6716;
  strcpy(depot.host, localibp_host);
#endif
  
  /* Open a new socket on a port */
  listening_socket = bindToFirstAvailablePort(3456,&my_port, 0);
  if (listening_socket == -1)
  {
#ifdef VIEW
    fprintf(stderr,"Impossible to bind to a port\n");
#endif
    if (sendInt(comm,-1) == -1)
      netsolvePerror(__FILE__,__LINE__);
    endTransaction(comm);
    notifyServiceFinished(global.my_self->host_desc->hostname,
                global.my_self->port,restriction_index);
    exit(0);
  }

  /* listening on the socket */
  listen(listening_socket,MAX_CONNECTIONS);

  /**************************************************/
  /* Send back my new port & pid                    */
  /**************************************************/

  my_pid = getpid();
  if ((sendInt(comm,my_port) == -1) ||
      (sendInt(comm,my_pid) == -1))
  {
    endTransaction(comm);
    notifyServiceFinished(global.my_self->host_desc->hostname,
                          global.my_self->port,restriction_index);
    cleanUp(tmpdir);
    exit(0);
  } 


  fprintf(stderr, "About to do the file stuffin generateservice_sequence.c\n");

  /*********************************************/
  /* Place myself in a new temporary directory */
  /*********************************************/

  while(1)
  {
    sprintf(tmpdir,"%s/netsolve-tmp-%d",global.scratch_path,(int)getpid()+count);
    if (!access(tmpdir,F_OK))
      count++;
    else
      break;
  }
  if (mkdir(tmpdir,0777))
  {
#ifdef VIEW
    fprintf(stderr,"Impossible to create directory %s\n",tmpdir);
#endif
    if (sendInt(comm,-1) == -1)
      netsolvePerror(__FILE__,__LINE__);
    endTransaction(comm);
    notifyServiceFinished(global.my_self->host_desc->hostname,
                global.my_self->port,restriction_index);
    exit(0);    
  } 

  if (chdir(tmpdir) == -1)
  {
#ifdef VIEW
    fprintf(stderr,"Impossible to chdir to %s\n",tmpdir);
#endif
    if (sendInt(comm,-1) == -1)
      netsolvePerror(__FILE__,__LINE__);
    endTransaction(comm);
    notifyServiceFinished(global.my_self->host_desc->hostname,
                          global.my_self->port,restriction_index);
    cleanUp(tmpdir);
    exit(0); 
  }


  /*****************************************/
  /* Save the problem descriptor to a file */
  /*****************************************/

  netsl_assign_file_names(top_node);  /* assigning file names for dependencies */

  fprintf(stderr, "Forall nodes: save pd\n");

  tmp_node = top_node;
  while (tmp_node != NULL) { /* for all nodes */
      sprintf(problemdescfile,"./pb_desc_%d",tmp_node->pd->job_id); /* one file for each node */
      fd = open(problemdescfile,O_WRONLY|O_CREAT,0666);
      if (fd < 0)
	  {
#ifdef VIEW
	      fprintf(stderr,"Impossible to create file '%s'\n",problemdescfile);
#endif
	      if (sendInt(comm,-1) == -1)
		  netsolvePerror(__FILE__,__LINE__);
	      endTransaction(comm);
	      notifyServiceFinished(global.my_self->host_desc->hostname,
				    global.my_self->port,restriction_index);
	      cleanUp(tmpdir);
	      exit(0);
	  }

      fprintf(stderr, "Writing pd for job_id %i\n", tmp_node->pd->job_id);
      
      if (writeProblemDescToFile(fd,tmp_node->pd) == -1)
	  {
#ifdef VIEW
	      fprintf(stderr,"Impossible to save problem descriptor to file %s\n",
		      problemdescfile);
#endif
	      if (sendInt(comm,-1) == -1)
		  netsolvePerror(__FILE__,__LINE__);
	      endTransaction(comm);
	      notifyServiceFinished(global.my_self->host_desc->hostname,
				    global.my_self->port,restriction_index);
	      cleanUp(tmpdir);
	      exit(0);
	  }
      close(fd);
      tmp_node = tmp_node->next;
  } /* for all nodes */

  /*******************************/
  /* Determine problem specifics */
  /*******************************/
  tmp_node = top_node;
  upf = 0;
  customized = 0;
  ftp = 0;

  fprintf(stderr, "Been there, done that, got the t-shirt\n");

  while (tmp_node != NULL){

      upf = upf || containsUPF(tmp_node->pd);
      customized = customized ||  strcmp(tmp_node->pd->customized,"NONE");
      ftp = ftp || tmp_node->pd->ftp;

      tmp_node = tmp_node -> next;
  }


  fprintf(stderr, "About to tell proxy about our port and pid\n");
  
  /* cut the connection to the client proxy */
  endTransaction(comm);

  /*********************************/
  /* Read all the objects to files */
  /*********************************/

  tmp_node = top_node;
  while (tmp_node != NULL){

  /* Wait for a connection from the client */
  sock = acceptConnection(listening_socket);
  if (sock == -1)
  {
#ifdef VIEW
    fprintf(stderr,"Error while accepting client connection\n");
#endif
    notifyServiceFinished(global.my_self->host_desc->hostname,
                          global.my_self->port,restriction_index);
    cleanUp(tmpdir);
    exit(0);
  }


  comm_to_client = acceptTransaction(sock);
  if (comm_to_client == NULL)
  {
#ifdef VIEW
    fprintf(stderr,"Error while accepting client transaction comm\n");
#endif
    notifyServiceFinished(global.my_self->host_desc->hostname,
                          global.my_self->port,restriction_index);
    cleanUp(tmpdir);
    exit(0);
  }

#ifdef KERBEROS5
  if (global.require_auth) {
    /* tell client we need credentials */
    tag = NS_PROT_KRB5_AUTH_REQUIRED;
    if (sendInt (comm_to_client, tag) == -1)
    {
      netsolvePerror (__FILE__,__LINE__);
      endTransaction(comm_to_client);
      notifyServiceFinished(global.my_self->host_desc->hostname,
                  global.my_self->port,restriction_index);
      exit(-1);
    }

    /* get credentials from client */
    if (RecvKerberos5Credentials (comm_to_client) < 0) {
      /*
       * if we didn't like the credentials; tell client that the
       * authentication failed, and this will be the error code
       * associated with the problem request.  otherwise, fall
       * through to other error checking below
       */
      tag = NS_PROT_AUTH_FAILED;
      if (sendInt (comm_to_client, tag) == -1)
      {
        netsolvePerror (__FILE__,__LINE__);
        endTransaction(comm_to_client);
        notifyServiceFinished(global.my_self->host_desc->hostname,
                  global.my_self->port,restriction_index);
        exit(-1);
      }
    }
    tag = NS_PROT_ACCEPTED;
    if (sendInt (comm_to_client, tag) == -1)
    {
      netsolvePerror (__FILE__,__LINE__);
      endTransaction(comm_to_client);
      notifyServiceFinished(global.my_self->host_desc->hostname,
                  global.my_self->port,restriction_index);
      exit(-1);
    }
  }
  else{
    tag = NS_PROT_ACCEPTED;
    if (sendInt (comm, tag) == -1)
    {
      netsolvePerror (__FILE__,__LINE__);
      endTransaction(comm_to_client);
      notifyServiceFinished(global.my_self->host_desc->hostname,
                  global.my_self->port,restriction_index);
      exit(-1);
    }
  }
#else
  tag = NS_PROT_ACCEPTED;
  if (sendInt (comm_to_client, tag) == -1)
  {
    netsolvePerror (__FILE__,__LINE__);
    endTransaction(comm_to_client);
    notifyServiceFinished(global.my_self->host_desc->hostname,
                global.my_self->port,restriction_index);
    exit(-1);
  }
#endif

  /* receiving the ID and the major */
  if ((recvInt(comm_to_client,&ID) == -1) ||
      (recvInt(comm_to_client,&client_major) == -1))
      {
#ifdef VIEW
	  fprintf(stderr,"Error while accepting client transaction\n");
#endif
	  /* Trying to send something back anyway !! */
	  if (sendInt(comm_to_client,NS_PROT_INTERNAL_FAILURE) == -1)
	      netsolvePerror(__FILE__,__LINE__);
	  endTransaction(comm_to_client);
	  notifyServiceFinished(global.my_self->host_desc->hostname,
				global.my_self->port,restriction_index);
	  /*	      cleanUp(tmpdir); */
	  exit(0);
      }
      
  /* Receiving the input objects */
#ifdef MULTINODESEQ
  if (recvInputObjectsToIBP_sequence(comm_to_client,&depot,tmp_node->pd) == -1)
#else
  if (recvInputObjectsToFiles_sequence(comm_to_client,tmp_node->pd) == -1)
#endif
	  {
#ifdef VIEW
	      fprintf(stderr,"Error while receiving input objects to files\n");
#endif
	      /* Trying to send something back anyway !! */
	      if (sendInt(comm_to_client,NS_PROT_INTERNAL_FAILURE) == -1)
		  netsolvePerror(__FILE__,__LINE__);
	      endTransaction(comm_to_client);
	      notifyServiceFinished(global.my_self->host_desc->hostname,
				    global.my_self->port,restriction_index);
	      /* cleanUp(tmpdir); */
	      exit(0);
	  }
      tmp_node = tmp_node -> next;
      endTransaction(comm_to_client);

  }
  /* Cutting the connection to the client */




  /* set all nodes to waiting */
  tmp_node = top_node;
  while (tmp_node !=NULL){
      tmp_node->exec_status = NS_DEPENDEND;
      tmp_node = tmp_node -> next;
  }

  time(&date1);


      if ((lifelink_pid = startLifeLink(proxy_IPaddr,proxy_port,NS_PROT_JOB_REPORT,
 				    ID,NetSolveServerError)) == -1)
      {
	  fprintf(stderr,"Warning: %s,%d: cannot start life link....\n",
		  __FILE__,__LINE__);
      } 
 


  /***************************************************/
  /* catching the signals again                      */
  /***************************************************/

#ifdef  SYSVSIGNAL
  (void)signal(SIGCLD,SIG_DFL);
#else
  (void)signal(SIGCHLD,SIG_DFL);
#endif 




  /***************************************************/
  /* Scheduler                                       */
  /***************************************************/

  while (1) {
    /*
     * Find out if we have any nodes that can run
     * (also sets the dependencies of all waiting processes correct)
     */
    if (netsl_may_run(top_node) == -1) {
/*        fprintf(stderr, */
/*  	      "netsl_may_run() on %s returned -1\n", */
/*  	      top_node ? top_node->pd->nickname : "null"); */
      break;
    }

    /*
     * Now fork all runnable nodes
     */
    if (netsl_fork_exec_all(top_node, listening_socket, ID, 
			    client_major, agent_name, restriction_index,
			    tmpdir, proxy_IPaddr, proxy_port) == -1) {
      fprintf(stderr, "netsl_fork_exec_all() returned -1 for top_node=\"%s\"\n",
	      top_node ? top_node->pd->nickname : "null");
      break;
    }

    /*
     * Wait until a child process exits - or until someone tells
     * us that they're done.
     */
    if(mspawns) {
      /*
       * We have multinode spawns... See if any completed...
       */
      fd_set readers;
      multinode_spawn_e * iter;
      int rc;
      int maxfd = 0;

      FD_ZERO(&readers);
      for(iter = mspawns; iter; iter = iter->next) {
	FD_SET(iter->remote->sock, &readers);
	maxfd = iter->remote->sock > maxfd ? iter->remote->sock : maxfd;
      }
      /*
       * Upon child completion, select() will return -1
       * and errno == EINTR.   See, that's a beautiful API  ;)
       */
      fprintf(stderr, "about to select ->maxfd=%i\n", maxfd);
      rc = select(maxfd+1, &readers, 0, 0, 0);
      fprintf(stderr, "select rc=%i errno=%i\n", rc, errno);
      if(rc == -1 && errno == EINTR) {
	/*
	 * See if any child process exited
	 */
	while(-1 == (wait_pid = waitpid(-1, &wait_status, WNOHANG)))
	  if(errno != EINTR) {
	    fprintf(stderr, "wait() failed! (%s)\n", strerror(errno));
	    break;
	  }

      } else if(rc == -1) {
	/*
	 * Seems like we caught an error...
	 */
	fprintf(stderr, "Caught error in select(): %s\n", strerror(errno));
	break;

      } else if(rc > 0) {
	/*
	 * Find the multinode spawn that exited
	 */
	multinode_spawn_e * piter = 0;
	for(iter = mspawns; iter; piter = iter, iter = iter->next)
	  if(FD_ISSET(iter->remote->sock, &readers))
	    break;

	if(!iter) {
	  /*
	   * We couldn't find it - can't happen !
	   */
	  fprintf(stderr, "select() returned an fd we didn't give it...\n");
	  break;
	}

	/*
	 * Ok - iter points us to the node that completed
	 */
	if(recvInt(iter->remote, &wait_status) == -1) {
	  fprintf(stderr, "Couldn't get status on remote execution\n");
	  wait_status = NS_PROT_INTERNAL_FAILURE;
	}

	/*
	 * Allocate a unique pid for the node and for our wait_pid
	 * variable.  This is done for re-use of netsl_resolve_dependencies().
	 */
	wait_pid = iter->node->process_id;

	/*
	 * End comms. deallocate, and unchain
	 */
	endTransaction(iter->remote);
	if(!piter)
	  mspawns = iter->next;
	else
	  piter->next = iter->next;
	free(iter);

      } else {
	/*
	 * select() returned 0.  It can't !
	 */
	fprintf(stderr, "select() doesn't seem to understand that it can't return 0.\n");
	break;
      }
    } else {

      /*
       * We have no active multinode spawns - just do as
       * we used to in the old days.
       */
      
      while(-1 == (wait_pid = wait(&wait_status)))
	if(errno != EINTR) {
	  fprintf(stderr, "wait() failed! (%s)\n", strerror(errno));
	  break;
	}

    }

    /*
     * (jakob) XXX: Catch WIFSIGNALED(wait_status) here, and set NS_errno
     * accordingly - pass back error to client program
     */

    /*
     * Now mark all dependencies of results from wait_pid as satisfied
     */
    if (netsl_resolve_dependencies(top_node, wait_pid, proxy_IPaddr, proxy_port) == -1) {
      fprintf(stderr, 
	      "netsl_resolve_dependencies() returned -1 for top_node \"%s\"\n",
	      top_node->pd->nickname);
      break;
    }
    
  }

  time(&date2);
  elapsed = (int)(date2-date1);

  netsl_exit_sequence_scheduler(top_node, lifelink_pid, tmpdir, proxy_IPaddr, 
				proxy_port, elapsed, NS_PROT_SOLVED, client_major, 
				restriction_index, ID, listening_socket );

  netsl_delete_Nodes(top_node);

  /* (jakob) XXX:
   * According to POSIX 3.3.1.3 setting the handler for SIGCHLD to SIG_IGN
   * is a very bad thing to do...  SYSV and BSD differ on behaviour in
   * this case - and by the way - why would we want to ignore the signal
   * at all ???
#ifdef  SYSVSIGNAL
    (void)signal(SIGCLD,SIG_IGN);
#else 
    (void)signal(SIGCHLD,SIG_IGN);
#endif 
   */

  exit (0);
}







/* send all needed information back to client and proxy */

int netsl_exit_sequence_scheduler(NS_Node *top_node, int lifelink_pid, char *wd, NS_IPaddr_type proxy_IPaddr, 
				  int proxy_port, int elapsed, int solved, int client_major, 
				  int restriction_index, int ID, int listening_socket ){
   NS_Socket_type   sock;
   NS_Communicator  *comm=NULL;
   NS_Node       *tmp_node;
   char stdoutfile[256];


 /* Notifying the proxy of completion */
   /* (jakob) XXX: The above comment is odd... What
    * we do here is we kill the lifelink process
    * with a SIGKILL, meaning; there is no chance
    * what so ever that it will ever be able to
    * actually contact the proxy.  But if it did,
    * it would send back an error to the proxy,
    * so it's ok that we SIGKILL it - the comment
    * above is just wrong...  Or are there deeper
    * reasons ?
    */
  kill(lifelink_pid,SIGKILL);

  /* Listening on the socket */
  if ((sock = acceptConnection(listening_socket)) == -1)
  {
#ifdef VIEW
    fprintf(stderr,"Error while accepting the connection.\n");
#endif
    notifyServiceFinished( global.my_self->host_desc->hostname, 
			   global.my_self->port, restriction_index);
    cleanUp(wd);
    exit(0);
  }
  comm = acceptTransaction(sock);
  if (comm == NULL)
  {
#ifdef VIEW
    fprintf(stderr,"Error while accepting the transaction.\n");
#endif
    notifyServiceFinished(global.my_self->host_desc->hostname, 
			   global.my_self->port, restriction_index);
    cleanUp(wd);
    exit(0);
  }

  /* Sending back the Computational time elapsed */
  if (sendInt(comm,elapsed) == -1)
  {
#ifdef VIEW
    fprintf(stderr,"Error while sending the elapsed time.\n");
#endif
    notifyServiceFinished(global.my_self->host_desc->hostname, 
			   global.my_self->port,restriction_index);
    endTransaction(comm);
    cleanUp(wd);
    exit(0);
  }

  /* Sending back the stdout */
  sprintf(stdoutfile,"%s/netsolve-stdout",wd);
  if (sendFileAsString(comm,stdoutfile) == -1)
  {
#ifdef VIEW
    fprintf(stderr,"Error while sending the stdout.\n");
#endif
    notifyServiceFinished(global.my_self->host_desc->hostname, 
			   global.my_self->port, restriction_index);
    endTransaction(comm);
    cleanUp(wd);
    exit(0);
  }

  /* Error */
  if (solved != NS_PROT_SOLVED)
  {
    sendInt(comm,solved);
    endTransaction(comm);
    notifyServiceFinished(global.my_self->host_desc->hostname, 
			   global.my_self->port ,restriction_index);
    cleanUp(wd);
    exit(0);
  }

  sendInt(comm,solved);

  /* Send back the output */
  /* this takes care of the transpose */
  tmp_node = top_node;
  while(tmp_node != NULL){
#ifdef MULTINODESEQ
    if (sendOutputObjectsFromIBP(comm,&depot,tmp_node->pd) == -1)
#else
      if (sendOutputObjectsFromFiles(comm,tmp_node->pd) == -1)
#endif
	  {
#ifdef VIEW
	      fprintf(stderr,"Error while sending the output objects.\n");
	      
#endif
	      fprintf(stderr,"Sleeping 1000 seconds\n");sleep(1000);
	      notifyServiceFinished(global.my_self->host_desc->hostname, 
			   global.my_self->port, restriction_index);
	  
	      endTransaction(comm);
	      cleanUp(wd);
	      exit(0);
	  }
      tmp_node = tmp_node -> next;
  } /* for all nodes */
  /* Terminating the transaction */
  endTransaction(comm);
  notifyServiceFinished(global.my_self->host_desc->hostname, 
			global.my_self->port, restriction_index);

#ifdef VIEW
  fprintf(stderr,"Completed\n");
#endif

  /* Cleaning up */
  close(listening_socket);
  cleanUp(wd);
  return 0;



} /*netsl_exit_agent() */



/* set the execution status of ode with wait_pid  to terminate */
/* update all other dependencies */

int netsl_resolve_dependencies(NS_Node *top_node, int wait_pid, NS_IPaddr_type proxy_IPaddr, int proxy_port) {

    NS_Node *tmp_node = top_node;
    int term_jobid = -1;
    int i;
    
    /* determine the node job_id of the terminated process */
    while (tmp_node != NULL){
	if (tmp_node->process_id == wait_pid) {
	  if(running_jobs == 0) {
	    fprintf(stderr, "BUG: Most odd, running_jobs is zero and a child exited...\n");
	  } else running_jobs--;
	  fprintf(stderr, "Procs: %i, running procs: %u\n",
		  global.my_self->host_desc->number_proc,
		  running_jobs);

	  if(running_jobs >= global.my_self->host_desc->number_proc) {
	    unsigned si;
	    fprintf(stderr, "Checking to see if we can use other server...\n");

	    for(si = 0; si < tmp_node->servers->nb_servers; si++)
	      if(tmp_node->servers->IPaddrlist[si] != global.my_self->host_desc->IPaddr)
		break;

	    if(si != tmp_node->servers->nb_servers) {
	      fprintf(stderr, "I want to use slave server %lu ...\n", 
		      (unsigned long)tmp_node->servers->IPaddrlist[si]);
	    }
	  }

	  fprintf(stderr, "CONFIRM: t=%li job %i is DONE\n", time(0), tmp_node->pd->job_id);

	  /*
	   * Set the exec_status to NS_TERMINATE
	   */
	    term_jobid = tmp_node->pd->job_id;
	    tmp_node->exec_status = NS_TERMINATE;
#ifdef SEQUENCING_DEBUG
	    fprintf(stderr, "Marked termination of %s\n", tmp_node->pd->nickname);
#endif
	    /*
	     * Now tell the proxy that the node terminated.
	     */
	    if(notifyProxy(proxy_IPaddr, proxy_port, NetSolvePartialCompletionOK, tmp_node->id, tmp_node->pd->job_id) == -1) {
	      /*
	       * Ugh !  We cannot tell the proxy that a problem in the sequence
	       * completed.  Might as well just abandon the entire computation then..
	       */
	      fprintf(stderr, "A problem completed but we couldn't notify the proxy. Aborting sequence!\n");
	      return -1;
	    }
	    fprintf(stderr, "Told proxy that job_id %i exited (with ID %i)\n", term_jobid, tmp_node->id);
	}
	tmp_node = tmp_node -> next;
    }
    if (term_jobid == -1) {
	fprintf(stderr,"Process not found:%d\n",wait_pid);
	return -1;
    }


    /* remove all dependencies from other nodes */
    tmp_node = top_node;
    while (tmp_node != NULL){ 
	for (i = 0; i < tmp_node->pd->nb_input_objects; i++){
	    if (tmp_node->in_dep->dest_node[i] == term_jobid){ /* input was dependent on terminated node */
		tmp_node->in_dep->dest_node[i] = -1;
		tmp_node->in_dep->dep_type[i] = NETSOLVE_NO_DEPENDENCY;
#ifdef SEQUENCING_DEBUG
		fprintf(stderr, "Resolved input dependency from %s input %i\n",
			tmp_node->pd->nickname, i);
#endif
	    }
	} /* for all input objects*/

	for (i = 0; i < tmp_node->pd->nb_output_objects; i++){
	    if (tmp_node->out_dep->dest_node[i] == term_jobid){ /* input was dependent on terminated node */
		tmp_node->out_dep->dest_node[i] = -1;
		tmp_node->out_dep->dep_type[i] = NETSOLVE_NO_DEPENDENCY;
#ifdef SEQUENCING_DEBUG
		fprintf(stderr, "Resolved output dependency from %s output %i\n",
			tmp_node->pd->nickname, i);
#endif
	    }
	} /* for all output objects */

#ifdef SEQUENCING_DEBUG
	fprintf(stderr, "---start total dependencies for %s----\n",
		tmp_node->pd->nickname);
	for (i = 0; i < tmp_node->pd->nb_input_objects; i++)
	  fprintf(stderr, " in %i - dep %i  with peer %i\n", 
		  i, tmp_node->in_dep->dep_type[i],
		  tmp_node->in_dep->dest_node[i]);
	for (i = 0; i < tmp_node->pd->nb_output_objects; i++)
	  fprintf(stderr, "out %i - dep %i  with peer %i\n",
		  i, tmp_node->out_dep->dep_type[i],
		  tmp_node->out_dep->dest_node[i]);	
	fprintf(stderr, "---end total dependencies for----\n");	
#endif

	tmp_node=tmp_node->next;
    } /* for all nodes */
    
    return 0;

}/* netsl_resolve_dependencies */

   /* check for processes that may run now */
int netsl_may_run (NS_Node *top_node){

    int i;
    NS_Node *tmp_node;
    int dep;
    tmp_node = top_node;
    
    while (tmp_node != NULL){
	dep = FALSE;
	if (tmp_node-> exec_status == NS_DEPENDEND){
	  /*
	   * Check dependencies on input objects
	   */
	    for (i = 0; i < tmp_node->pd->nb_input_objects; i++){
		if ((tmp_node->in_dep->dep_type[i] == NETSOLVE_INPUT_DEPENDENCY) ||
		    (tmp_node->in_dep->dep_type[i] == NETSOLVE_OUTPUT_DEPENDENCY)){
		    dep = TRUE;
#ifdef SEQUENCING_DEBUG
		    fprintf(stderr, 
			    "netsl_may_run() got early dep=TRUE (%d) for %s i=%d "
			    "and top node %s\n",
			    tmp_node->in_dep->dep_type[i],
			    tmp_node->pd->nickname, 
			    i, top_node->pd->nickname);
#endif
		}
	    }
	    if (dep == FALSE){
	      /*
	       * Check dependencies on output objects
	       */
		for (i = 0; i < tmp_node->pd->nb_output_objects; i++){
		  if (((tmp_node->out_dep->dep_type[i] == NETSOLVE_INPUT_DEPENDENCY) && 
		       (tmp_node->out_dep->dest_node[i] != tmp_node->pd->job_id ))
/*  		       (tmp_node->out_dep->dest_node[i] != i )) */
		      ||     /* no local dependency */
		      (tmp_node->out_dep->dep_type[i] == NETSOLVE_OUTPUT_DEPENDENCY)){
		    dep = TRUE;
#ifdef SEQUENCING_DEBUG
		    fprintf(stderr,
			    "netsl_may_run() got late dep=TRUE (dep_type=%d) "
			    "for %s out i=%d "
			    "and top node %s\n",
			    tmp_node->out_dep->dep_type[i],
			    tmp_node->pd->nickname, 
			    i, top_node->pd->nickname);
#endif
		  }
		}
		/*
		 * If there are no dependencies stopping the node from
		 * executing, allow it to execute
		 */
		if (dep == FALSE){	    
		    tmp_node->exec_status = NS_WAITING;
#ifdef SEQUENCING_DEBUG
		    fprintf(stderr, "netsl_may_run() set NS_WAITING for %s\n",
			    tmp_node->pd->nickname);
#endif
		}
	    }  /*if */
	}  /*if DEPENDEND */
	tmp_node=tmp_node->next;
    } /* for all nodes */

    /*
     * We must return -1 if no nodes are runnable
     */

    tmp_node = top_node;
    while(tmp_node != NULL){
      if (tmp_node->exec_status != NS_TERMINATE)
	return 0;
/*        if(tmp_node->exec_status == NS_WAITING) */
/*  	return 0; */
      tmp_node = tmp_node->next;
    }


    return -1;
} /* netsl_may run */


/*
 * This routine requests that a node be spawned on a
 * remote server.
 */
int remote_netsl_fork_exec(NS_Node * node)
{
  int remsrv;
  NS_Communicator * remote;
  int reply;
  multinode_spawn_e * splist;

  /*
   * Pick a remote server randomly from the list of remote
   * servers in the nodes server list
   */
  
  /* Fail if we're the only server */
  if(node->servers->nb_servers <= 1) 
    return 0;
  /* Quick, if there's only one more */
  if(node->servers->nb_servers == 2) {
    if(node->servers->IPaddrlist[0] == global.my_self->host_desc->IPaddr)
      remsrv = 1;
    else
      remsrv = 0;
  } else {
    /*
     * Ok, we can't do it quickly...
     */
    do {
      static lastpick = 0;
      lastpick = remsrv = (lastpick+1) % node->servers->nb_servers;
    } while(node->servers->IPaddrlist[remsrv] == global.my_self->host_desc->IPaddr);

fprintf(stderr, "SLOWSELECT picked remote server # %i out of %i\n", remsrv, node->servers->nb_servers);

  }

  /*
   * Good, now we have a remote server that is not ourselves  ;)
   */

  /*
   * reset process_id. - we'll make one up when we know the
   * job is done....
   */  
  node->process_id = 0;

  /*
   * Connect to server
   */
  remote = contactHost(0, node->servers->IPaddrlist[remsrv], 
		       node->servers->portlist[remsrv], DATA_XDR);
  if(!remote) {
    /*
     * No harm done...
     */
    fprintf(stderr, "Couldn't contact remote server - spawning locally\n");
    return 0;
  }

  /*
   * Send protocol request
   */
  if(sendInt(remote, NS_PROT_MULTINODE_SPAWN) == -1) {
    fprintf(stderr, "Failed sending multinode spawn request to peer\n");
    endTransaction(remote);
    return 0;
  }

  /*
   * See if we were accepted...
   */
  if(recvInt(remote, &reply) == -1) {
    fprintf(stderr, "Failed reading protocol initiation reply\n");
    endTransaction(remote);
    return 0;
  }
  if(reply != NS_PROT_OK) {
    fprintf(stderr, "Peer didn't want to talk to us - probably doesn't support multinode\n");
    endTransaction(remote);
    return 0;
  }

  /*
   * Send problem description
   */
  if(sendProblemDesc(remote, node->pd) == -1) {
    fprintf(stderr, "Error sending problem description to remote node\n");
    endTransaction(remote);
    return 0;
  } 

  /* 
   * Read response
   */
  if(recvInt(remote, &reply) == -1) {
    fprintf(stderr, "Failed reading multinode spawn reply\n");
    endTransaction(remote);
    return 0;
  }
  if(reply != NS_PROT_OK) {
    fprintf(stderr, "Peer rejected the spawn request\n");
    endTransaction(remote);
    return 0;
  }
  
  fprintf(stderr, "Job successfully spawned on remote node.\n");

  /*
   * Put connection in list of connections to keep an eye on
   */
  splist = (multinode_spawn_e*)malloc(sizeof(multinode_spawn_e));
  if(!splist) {
    /*
     * Let's pretend nothing happened - the remote node will
     * do the computation and we won't care - but nothing bad will
     * (probably) happen.
     */
    fprintf(stderr, "Hey - no more meory !\n");
    return 0;
  }
  splist->node = node;
  splist->remote = remote;
  splist->next = mspawns;
  mspawns = splist;

  /* success */
  return 1;
}

/* fork off a new process for each waiting job */
/* store the pid to determine, process termination */

int netsl_fork_exec_all (NS_Node *top_node, int listening_socket, int ID, 
			 int client_major, char* agent_name, int restriction_index,
			 char* tmpdir, int proxy_IPaddr, int proxy_port){

    NS_Node *tmp_node;
    char *xname;
    char *xpath;
    char buffer[256];
    char *arch;
    int lifelink_pid = 0;
    int  child_pid;
    

    tmp_node = top_node;
    while (tmp_node != NULL){
	/* may this node run? */
	if (tmp_node->exec_status == NS_WAITING) { /* this node whises to be executed */
	    tmp_node->exec_status = NS_RUNNING; /* set status to running */

	    /***************************************************/
	    /* Set the default path and name of the executable */
	    /***************************************************/
	    arch = getNetSolveArch();
	    sprintf(buffer,"%s/bin/%s/",
		    global.netsolve_root_path,arch);
	    free(arch);
	    xpath = strdup(buffer);
	    sprintf(buffer,"service-%s",tmp_node->pd->file);
	    xname = strdup(buffer);


	    fprintf(stderr, "Procs: %i, running procs: %u\n",
		    global.my_self->host_desc->number_proc,
		    running_jobs);

	    /*
	     * If running_jobs >= global.my_self->host_desc->number_proc,
	     * then we have at least one running service per processor.
	     * We should then consider spawning the next node on a
	     * remote server...  
	     * Well, the node contains a list of servers on which it
	     * can run, and we're guaranteed that the server list only
	     * includes servers in our IBP group.  Just pick one and
	     * get it over with...
	     * (XXX: Ideally we would be asking the agent which in turn would
	     *  know about who's doing what and where)
	     */
	    if(global.my_self->host_desc->number_proc <= running_jobs) {

	      fprintf(stderr, "Ok - we should try to spawn on remote node\n");
	      
	      /*
	       * If remote spawn succeeds, return.  Otherwise, keep it local then.
	       */
	      if(remote_netsl_fork_exec(tmp_node)) {
		int tmp_id;
		/* Ok - return */
		fprintf(stderr, "I believe we successfully forked off a job "
			"to a remote multinode server\n");

		/*
		 * Make up a process ID we know others won't use
		 * (a negative one)
		 */
		do {
		  NS_Node * niter;
		  tmp_id = -(rand() % 0xFFFF);
		  for(niter = top_node; niter; niter = niter->next)
		    if(niter->process_id == tmp_id)
		      tmp_id = 0;
		} while(!tmp_id);
		tmp_node->process_id = tmp_id;

		tmp_node = tmp_node->next;
		continue;		
	      }     
	    } 



#ifdef SEQUENCING_DEBUG
	    fprintf(stderr, "Ok, we're right about to fork off node id %i (job_id %i)\n",
		    tmp_node->id, tmp_node->pd->job_id);
#endif

	    child_pid =  fork();

            if (child_pid == 0){


		/*************************************************************/
		/* Start the life link to the proxy                          */
		/*************************************************************/
	  fprintf(stderr, "ALERT: t=%li job %i is START\n",
		  time(0), tmp_node->pd->job_id);

	      forkStandardFileService_sequence(listening_socket,ID,client_major,
					       lifelink_pid,agent_name,tmpdir, 
					       global.my_self->host_desc->hostname,
					       global.my_self->port,restriction_index, 
					       xpath,xname,tmp_node->pd,proxy_IPaddr,proxy_port,
					       tmp_node->id, depot.host); 
		printf("exiting:%d\n",getpid());	       
		exit (0);
	    }
	    else{
		tmp_node->process_id  = child_pid;
		running_jobs++;
		fprintf(stderr, "Procs: %i, running procs: %u\n",
			global.my_self->host_desc->number_proc,
			running_jobs);
	    }

	}

    tmp_node = tmp_node-> next;
    }

  return 0;
}

/* recursive resolve links */
/* follow all links rekursive and assign the filename of hte first occurence */
#ifdef MULTINODESEQ
struct ibp_set_of_caps *netsl_resolve_name(NS_Node *act_node, int dep_type, 
			 int dest_node_job_id, int dest_arg)
#else
char *netsl_resolve_name(NS_Node *act_node, int dep_type, 
			 int dest_node_job_id, int dest_arg)
#endif
{
#ifdef MULTINODESEQ
  struct ibp_depot depot;
  struct ibp_attributes attrib;
  static const char * ibp_server_name = 0;

  if(!ibp_server_name)
    ibp_server_name = reverse_lookup(act_node->servers->IBPaddrlist + 0);
  
  if(!ibp_server_name) {
    fprintf(stderr, "Hey, can't resolve name of IBP server !\n");
    return 0;
  }
  strcpy(depot.host, ibp_server_name);
  depot.port = 6716; /* default port */
  
  attrib.duration = 0;
  attrib.reliability = IBP_STABLE;
  attrib.type = IBP_BYTEARRAY;
#else
  char *buffer;
  char * r_buffer;

  buffer = malloc(100 * sizeof (char));
#endif

  while(act_node && act_node->pd->job_id != dest_node_job_id) {
    act_node = act_node->last; /* find the dest_node */
  }

  if(!act_node) {
    fprintf(stderr, "Hey ! act_node not found for job_id %i (pid %i)\n", dest_node_job_id, getpid());
    return 0;
  }

    switch (dep_type) {
    case NETSOLVE_INPUT_DEPENDENCY:
	if (act_node->in_dep->dep_type[dest_arg] == NETSOLVE_NO_DEPENDENCY) {
#ifdef MULTINODESEQ
	  IBP_set_of_caps caps;

	  /*
	   * We simply copy the capabilities from the output argument we depend upon
	   */
	  caps = (struct ibp_set_of_caps*)malloc(sizeof(struct ibp_set_of_caps));
	  caps->readCap = strdup(act_node->pd->input_names[dest_arg]->readCap);
	  caps->writeCap = strdup(act_node->pd->input_names[dest_arg]->writeCap);
	  caps->manageCap = strdup(act_node->pd->input_names[dest_arg]->manageCap);
	  fprintf(stderr, "CAPDUP: pid=%i, duplicated indep arg=%i, cap=\"%s\"\n",
		  getpid(), dest_arg, caps->readCap);

	  return caps;	  
#else
	  sprintf(buffer,"input%d_%d", act_node->pd->job_id, dest_arg);
	  r_buffer = strdup(buffer);
	  free (buffer);

	  return r_buffer;
#endif
	} else {
#ifndef MULTINODESEQ
	  free (buffer);
#endif
	  fprintf(stderr, "CAPDUP: Extra recursion with INDEP\n");
	  return netsl_resolve_name(act_node, act_node->in_dep->dep_type[dest_arg],  
				    act_node->in_dep->dest_node[dest_arg],  
				    act_node->in_dep->dest_arg[dest_arg]);
	}	    
	break;
    case NETSOLVE_OUTPUT_DEPENDENCY:


#ifdef MULTINODESEQ
      /*
       * Ok, in the case of OUTPUT dependencies, a special rule applies when
       * we're using IBP...
       * Since IBP doesn't support overwriting of files, we have allocated a
       * separate output for each of the chain of dependant outputs in the graph.
       * Therefore we do not wish to follow this dependency all the way up,
       * will should instead use the first node we depend upon.
       */
      if(1) {
	  IBP_set_of_caps caps;
	  caps = (struct ibp_set_of_caps*)malloc(sizeof(struct ibp_set_of_caps));
	  caps->readCap = strdup(act_node->pd->output_names[dest_arg]->readCap);
	  caps->writeCap = strdup(act_node->pd->output_names[dest_arg]->writeCap);
	  caps->manageCap = strdup(act_node->pd->output_names[dest_arg]->manageCap);
	  fprintf(stderr, "CAPDUP: pid=%i, duplicated outdep DIRECT arg=%i, cap=\"%s\"\n",
		  getpid(), dest_arg, caps->readCap);
	  return caps;
      }

#endif


	if (act_node->out_dep->dep_type[dest_arg] == NETSOLVE_NO_DEPENDENCY){
#ifdef MULTINODESEQ
	  IBP_set_of_caps caps;

#if 0
	  caps = IBP_allocate(&depot, 0, act_node->output_size * 2, &attrib);
	  if(!caps) {
	    fprintf(stderr, "Error getting IBP capability! -> B %i\n", IBP_errno);
	    return 0;
	  }
#else
	  caps = (struct ibp_set_of_caps*)malloc(sizeof(struct ibp_set_of_caps));
	  caps->readCap = strdup(act_node->pd->output_names[dest_arg]->readCap);
	  caps->writeCap = strdup(act_node->pd->output_names[dest_arg]->writeCap);
	  caps->manageCap = strdup(act_node->pd->output_names[dest_arg]->manageCap);
	  fprintf(stderr, "CAPDUP: pid=%i, duplicated outdep arg=%i, cap=\"%s\"\n",
		  getpid(), dest_arg, caps->readCap);
#endif

	  return caps;
#else
	  sprintf(buffer,"output%d_%d", act_node->pd->job_id, dest_arg);
	  r_buffer = strdup(buffer);
	  free (buffer);
	  return r_buffer;
#endif
	} else {
#ifndef MULTINODESEQ
	  free (buffer);
#endif
	  fprintf(stderr, "CAPDUP: Extra recursion with OUTDEP\n");
	  return netsl_resolve_name(act_node, act_node->out_dep->dep_type[dest_arg],  
				    act_node->out_dep->dest_node[dest_arg],  
				    act_node->out_dep->dest_arg[dest_arg]);
	}	
	break;
    }


    return 0;
} /* char *netsl_resolve_name(NS_Node *act_node, int dep_type, int dest_node_job_id, int dest_arg) */



/* assigns a filename to each input and output */
int netsl_assign_file_names (NS_Node *top_node){

    int i;
    NS_Node *tmp_node = top_node;

#ifdef MULTINODESEQ
    /*
     * Allocate names from IBP
     */
    struct ibp_depot depot;
    struct ibp_attributes attrib;
    const char * ibp_server_name = reverse_lookup(top_node->servers->IBPaddrlist + 0);

    if(!ibp_server_name) {
      fprintf(stderr, "Hey, can't resolve name of IBP server !\n");
      return -1;
    }
    strcpy(depot.host, ibp_server_name);
    depot.port = 6716; /* default port */

    attrib.duration = 0;
    attrib.reliability = IBP_STABLE;
    attrib.type = IBP_BYTEARRAY;

    while(tmp_node) {
      /* Make sure name array is allocated */
      if(!tmp_node->pd->input_names)
	tmp_node->pd->input_names = (struct ibp_set_of_caps**)malloc(tmp_node->pd->nb_input_objects 
								     * sizeof(struct ibp_set_of_caps*));
      if(!tmp_node->pd->input_names) {
	fprintf(stderr, "Memory allocation error for input name array\n");
	return -1;
      }

      /*
       * For each input object, allocate an IBP capability
       */
      for(i = 0; i< tmp_node->pd->nb_input_objects; i++) {

	if(tmp_node->in_dep->dep_type[i] == NETSOLVE_NO_DEPENDENCY) {
	  /* No dependenies - just get a name */
	  IBP_set_of_caps caps;

	  caps = IBP_allocate(&depot, 0, tmp_node->input_size * 2, &attrib);
	  if(!caps) {
	    fprintf(stderr, "Error getting IBP capability! -> C %i\n", IBP_errno);
	    return -1;
	  }

	  fprintf(stderr, "CAPRES: pid=%i, node=\"%s\", in_arg=%i, readCap=\"%s\"\n",
		  getpid(), tmp_node->pd->nickname, i, caps->readCap);

	  tmp_node->pd->input_names[i] = caps;

	} else {
	  /* recursive link trace */
	  tmp_node->pd->input_names[i] = netsl_resolve_name(tmp_node, tmp_node->in_dep->dep_type[i],  
							    tmp_node->in_dep->dest_node[i],  
							    tmp_node->in_dep->dest_arg[i]);
	  fprintf(stderr, "CAPRECUR: pid=%i, node=\"%s\", arg=%i, readCap=\"%s\"\n",
		  getpid(), tmp_node->pd->nickname, i, tmp_node->pd->input_names[i]->readCap);
	}
      }
      /* Make sure output_names array is allocated */
      if(!tmp_node->pd->output_names) 
	tmp_node->pd->output_names = (struct ibp_set_of_caps**)malloc(tmp_node->pd->nb_output_objects
								      * sizeof(struct ibp_set_of_caps*));
      if(!tmp_node->pd->output_names) {
	fprintf(stderr, "Out of memory allocating output_names array\n");
	return -1;
      }

      /*
       * For each output, allocate an IBP capability set
       */      
      for(i = 0; i< tmp_node->pd->nb_output_objects; i++){
	/* Direct ? */


	/* ALWAYS ALLOCATE OUTPUTS... */

#if 0
	if (tmp_node->out_dep->dep_type[i] == NETSOLVE_NO_DEPENDENCY) {
#endif

	  IBP_set_of_caps caps;

	  /*
	   * We use the input size as basis since the output_size is meaningless (see comment in determinesizes.c)
	   * so input_size has a better chance of making sense, although all this sizing stuff is pretty
	   * far fetched...
	   */

	  caps = IBP_allocate(&depot, 0, 100 + tmp_node->input_size * 2, &attrib);
	  if(!caps) {
	    fprintf(stderr, "Error getting IBP capability for output! -> D %i\n", IBP_errno);
	    return -1;
	  }
	  tmp_node->pd->output_names[i] = caps;

	  fprintf(stderr, "CAPALLOC: pid=%i, size=%i, out_arg=%i, readCap=\"%s\"\n",
		  getpid(), 100 + tmp_node->input_size * 2, i, caps->readCap);

#if 0
	} else {

	  /* recursive link trace */
	  tmp_node->pd->output_names[i] = netsl_resolve_name(tmp_node, tmp_node->out_dep->dep_type[i],  
							     tmp_node->out_dep->dest_node[i],  
							     tmp_node->out_dep->dest_arg[i]);
	  fprintf(stderr, "CAPRECUR: pid=%i, node=\"%s\", out_arg=%i, readCap=\"%s\"\n",
		  getpid(), tmp_node->pd->nickname, i, tmp_node->pd->output_names[i]->readCap);
	}
#endif

      }
      
      tmp_node = tmp_node-> next;
    }

#else
    char *buffer;
    buffer = malloc(100 * sizeof (char));

    while(tmp_node != NULL){
      if (tmp_node->pd->input_names == NULL) 
	  tmp_node->pd->input_names = malloc ( tmp_node->pd->nb_input_objects * sizeof (char *));
	for (i = 0; i< tmp_node->pd->nb_input_objects; i++){
	    if (tmp_node->in_dep->dep_type[i] == NETSOLVE_NO_DEPENDENCY){
		sprintf(buffer,"input%d_%d", tmp_node->pd->job_id,i); /* no link */
		tmp_node->pd->input_names[i] = strdup(buffer);
	    }
	    else /* recursive link trace */{
		tmp_node->pd->input_names[i] = netsl_resolve_name(tmp_node, tmp_node->in_dep->dep_type[i],  
								  tmp_node->in_dep->dest_node[i],  
								  tmp_node->in_dep->dest_arg[i]);
	    }


	}
      if (tmp_node->pd->output_names == NULL) 
	  tmp_node->pd->output_names = malloc ( tmp_node->pd->nb_output_objects * sizeof (char *));

	for (i = 0; i< tmp_node->pd->nb_output_objects; i++){
	    if (tmp_node->out_dep->dep_type[i] == NETSOLVE_NO_DEPENDENCY){
		sprintf(buffer,"output%d_%d", tmp_node->pd->job_id,i); /* no link */
		tmp_node->pd->output_names[i]=strdup(buffer);
	    }
	    else{ /* recursive link trace */
		tmp_node->pd->output_names[i] = netsl_resolve_name(tmp_node, tmp_node->out_dep->dep_type[i],  
								  tmp_node->out_dep->dest_node[i],  
								  tmp_node->out_dep->dest_arg[i]);
	    }
	}
	
	tmp_node = tmp_node-> next;
    }

    free(buffer);
#endif

    return 0;

} /* netsl_assign_file_names */




