/*****************************************************************/
/*      netsolveupf.c                                            */
/*      Henri Casanova                                           */
/*****************************************************************/

#ifndef MULTINODESEQ

#include "core.h"
#include "netsolveupf.h"
#include "serverglobal.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

static char *safe_calls[] = {
    "D_SIGN",
    "TAN",
    "TANF",
    "TANH",
    "TANHF",
    "SINH",
    "SINHF",
    "COSH",
    "COSHF",
    "ATAN",
    "ATANF",
    "COS",
    "COSF",
    "SIN",
    "SINF",
    "LOG",
    "LOGF",
    "SQRT",
    "SQRTF",
    "EXP",
    "EXPF",
    "S_STOP",
    "STRCPY",
    "STRCMP",
    "STRNCMP",
    "STRNCPY",
    "MALLOC",
    "FREE",
    ""
};

/*
 * processUPF()
 */
int processUPF(NS_ProblemDesc *pd,char *wd,char **xname,char **xpath)
{
  int i;
  char *filename;
  int language;
  int status;
  char buffer[256];
  int fd;

  /* generating the .o files */
  for (i=0;i<pd->nb_input_objects;i++)
  {
    if (pd->input_objects[i]->object_type != NETSOLVE_UPF)
      continue;
    sprintf(buffer,"upfinput%d",i);
    pd->input_objects[i]->attributes.upf_attributes.filename = NULL;
    pd->input_objects[i]->attributes.upf_attributes.funcname = NULL;
    pd->input_objects[i]->attributes.upf_attributes.language = -1;
    sprintf(buffer,"./input%d",i);
    fd = open(buffer,O_RDONLY,0666);
    if (fd < 0)
    {
#ifdef VIEW
      fprintf(stderr,"File '%s' not found\n",buffer);
#endif
      ns_errno = NetSolveFileError;
      return -1;
    }
    if (readObjectFromFile(fd,pd->input_objects[i]) == -1)
      return -1;
    close(fd);
    filename = pd->input_objects[i]->attributes.upf_attributes.filename;
    language = pd->input_objects[i]->attributes.upf_attributes.language;
    
    status = generateUPFobject(filename,language);
    if (status == -1)
      return -1;
  }

  /* Recompile the numerical-x file */
  if (reCompileNumerical(pd,wd) == -1)
    return -1;

  /* Perform the link */
   if (reLink(pd,wd,xname,xpath) == -1)
     return -1;

   return 1;
}

/*
 * reCompileNumerical
 */
int reCompileNumerical(NS_ProblemDesc *pd,char *wd)
{
  char command[256];
  char makefilename[256];
  char buffer[256];
  FILE *f;
  struct stat st;
  int i;
  char *arch;

  f = fopen("./effectiveupf.h","w");
  if (f == NULL)
  {
#ifdef VIEW
    fprintf(stderr,"Impossible to create file 'effectiveupf.h'\n");
#endif
    ns_errno = NetSolveFileError;
    return -1;
  }
  for (i=0;i<pd->nb_input_objects;i++)
  { 
    if (pd->input_objects[i]->object_type != NETSOLVE_UPF)
      continue;
#ifdef ADD_
    fprintf(f,"#define upf%d %s_\n",i,
        pd->input_objects[i]->attributes.upf_attributes.funcname);
#endif
#ifdef NOCHANGE
    fprintf(f,"#define upf%d %s\n",i,
        pd->input_objects[i]->attributes.upf_attributes.funcname);
#endif
#ifdef UPCASE
    netsolveCaps(pd->input_objects[i]->attributes.upf_attributes.funcname);
    fprintf(f,"#define upf%d %s\n",i,
        pd->input_objects[i]->attributes.upf_attributes.funcname);
#endif
#ifdef F2C
    fprintf(f,"#define upf%d %s__\n",i,
        pd->input_objects[i]->attributes.upf_attributes.funcname);
#endif
  }
  fclose(f);

  sprintf(makefilename,"Makefile.numerical-%s",pd->nickname);
  f = fopen(makefilename,"w");
  if (f == NULL)
  {
#ifdef VIEW
    fprintf(stderr,"Impossible to create file '%s'\n",makefilename);
#endif
    ns_errno = NetSolveFileError;
    return -1;
  }
  arch = getNetSolveArch();
  fprintf(f,"include %s/conf/%s.def\n",
          global.netsolve_root_path,arch);
  free(arch);
  fprintf(f,"include %s/src/Makefile.def\n\n",
          global.netsolve_root_path);
  fprintf(f,"numerical-%s.o:\n",pd->file);
  fprintf(f,"\t$(CC) $(CFLAGS) -I. -c $(NETSOLVE_ROOT)/src/Server/numerical-%s.c -o $@\n",
          pd->file);
  fclose(f);

  sprintf(command,"make -f %s",makefilename);
  system(command);

  /* Has it worked ? */
  sprintf(buffer,"./numerical-%s.o",pd->file);
  if (stat(buffer,&st))
  {
#ifdef VIEW
    fprintf(stderr,"Compilation failed\n");
#endif
    ns_errno = NetSolveUPFError;
    return -1;
  }

  return 1;
}


/*
 * generateUPFobject()
 */
int generateUPFobject(char *filename, int language)
{
  char command[256];
  char buffer[256];
  char *makefilename;
  struct stat st;
  int safe;

  /* rename the file according to the language */
  switch(language)
  {
    case UPF_LANG_FORTRAN:
      sprintf(command,"mv %s %s.f",filename,filename);
      break;
    case UPF_LANG_C:
      sprintf(command,"mv %s %s.c",filename,filename);
      break;
    default:
      ns_errno = NetSolveInternalError;
      return -1;
  } 

  system(command);

  /* Create a small Makefile */
  makefilename = createUPFMakefile(filename,language);
  if (makefilename == NULL)
    return -1;
  sprintf(buffer,"make -f %s",makefilename);
  system(buffer);
  free(makefilename);

  /* Check that the object file has been generated */
  sprintf(buffer,"./%s.o",filename);
  if (stat(buffer,&st))
  {
#ifdef VIEW
    fprintf(stderr,"Object file '%s' not generated\n",buffer);
#endif
    ns_errno = NetSolveUPFError;
    return -1; 
  }
 
  /* Check the security */
  safe = checkSecurity(buffer);
  if (safe == -1)
    return -1;
  if (safe == 0)
  {
#ifdef VIEW
    fprintf(stderr,"UPF appears to be unsafe\n");
#endif
    ns_errno = NetSolveUPFUnsafe;
    return -1;
  }
  return 1;
}

/*
 * createUPFMakefile()
 */
char *createUPFMakefile(char *filename,int language)
{
  char buffer[256];
  char *makefilename;
  FILE *f;
  char *arch;

  sprintf(buffer,"./Makefile.%s",filename);
  makefilename = strdup(buffer);

  f = fopen(buffer,"w");
  if (f == NULL)
  {
#ifdef VIEW
    fprintf(stderr,"Impossible to create file '%s'\n",makefilename);
#endif
    free(makefilename);
    ns_errno = NetSolveFileError;
    return NULL;
  }

  arch = getNetSolveArch();
  fprintf(f,"include %s/conf/%s.def\n\n",global.netsolve_root_path,
           arch);
  free(arch);

  fprintf(f,"default: %s.o\n\n",filename);
  fprintf(f,"%s.o : %s",filename,filename);
  switch(language)
  {
    case UPF_LANG_FORTRAN:
      fprintf(f,".f\n");
      fprintf(f,"\t$(F77) $(F77DASHC) %s.f $(F77DASHO) %s.o\n",filename,filename);
      break;
    case UPF_LANG_C:
      fprintf(f,".c\n");
      fprintf(f,"\t$(CC) -c %s.f -o %s.o\n",filename,filename);
      break;
    default:
      ns_errno = NetSolveInternalError;
      return NULL;
  }
  fclose(f);
  return makefilename;
}

/*
 * checkSecurity()
 */
int checkSecurity(char *object)
{
  FILE *stream;
  char command[256];
  char line[256];

  sprintf(command,"nm -u %s",object);
  stream = popen(command,"r");
  if (stream == NULL)
  {
    perror("popen");
    ns_errno = NetSolveSystemError;
    return -1;
  }

  while(fgets(line,256,stream) != NULL)
  {
    char *begin = line;
    char *end;

#ifdef PLAT_SUN4SOL2
    if (strncmp(line,"Undefined symbols from",22) == 0)
      continue;
#endif
#ifdef PLAT_T3E
    if (strncmp(line,"./",2) == 0)
      continue;
#endif
#ifdef PLAT_SGI64
    if (strncmp(line,"Undefined",9) == 0)
      continue;
#endif
#ifdef PLAT_ALPHA
    if (strncmp(line,"Name        ",12) == 0)
      continue;
#endif
    if (strcmp(line,"\n") == 0)
      continue;
 
    while((*begin == '_') || (*begin == ' '))
      begin++;

    end = (char *)((char *)line + strlen(line))-1;
    while ((*end == '\n')||(*end == ' ')||(*end == '_'))
      *(end--)='\0';

#ifdef PLAT_SGI64
    begin = end;
    while ((*(begin-1) != '|'))
      begin--;
#endif

    netsolveCaps(begin);
    if (!isAuthorized(begin))
      return 0;
  }
  pclose(stream);
  return 1;
}

/*
 * isAuthorized()
 */
int isAuthorized(char *s)
{
  int i=0;

  while (safe_calls[i][0] != '\0')
  {
    if (!strcmp(s,safe_calls[i]))
      return 1;
    i++;
  }
#ifdef VIEW
  fprintf(stderr,"'%s' is an unauthorized symbol\n",s);
#endif
  return 0;
}

/*
 *  reLink()
 */
int reLink(NS_ProblemDesc *pd,char*wd,char **xname,char **xpath)
{
  char buffer[256];
  char makefilename[256];
  char command[256];
  FILE *f;
  struct stat st;
  int i;
  char *arch;

  sprintf(makefilename,"Makefile.service-%s",pd->nickname);
  f = fopen(makefilename,"w");
  if (f == NULL)
  {
#ifdef VIEW
    fprintf(stderr,"Impossible to create file '%s'\n",makefilename);
#endif
    ns_errno = NetSolveFileError;
    return -1;
  }
  arch = getNetSolveArch();
  fprintf(f,"default: %s/service-%s-upfed\n\n",wd,pd->file);
  fprintf(f,"include %s/conf/%s.def\n",global.netsolve_root_path,
                       arch);
  free(arch);
  fprintf(f,"include %s/src/Makefile.def\n\n",global.netsolve_root_path);
  fprintf(f,"include %s/src/Makefile.num_libs\n",global.netsolve_root_path);
  
  fprintf(f,"%s/service-%s-upfed : \n",wd,pd->file);
  fprintf(f,"\t cd $(OBJDIR);$(LINKER) $(UPF) \\\n");
  for (i=0;i<pd->nb_input_objects;i++)
  {
    if (pd->input_objects[i]->object_type != NETSOLVE_UPF)
      continue;
    fprintf(f,"\t\t%s/input%d-upf.o\\\n",wd,i);
  }
  fprintf(f,"\t\t%s/numerical-%s.o \\\n",wd,pd->file);
  fprintf(f,"\t\t$(F_FROM_C_LIB) \\\n");
  if (!strcmp(pd->customized,"NONE"))
  {
    fprintf(f,"\t\t$(STANDARDSERVICEOBJ) \\\n");
  }
  else if (!strcmp(pd->customized,"PETSC") )
  {
    fprintf(f," PETSC: TO DO\n");
  }
  else if (!strcmp(pd->customized,"SCALAPACK_MPI"))
  {
    fprintf(f," SCALAPACK_MPI: TO DO\n");
  }
  else if (!strcmp(pd->customized,"CONDOR"))
  {
    fprintf(f," CONDOR: TO DO\n");
  }
  fprintf(f,"\t\t$(LIBS) $(HBMLIB) $(ARCHDLIB) $(AUTH_LIBS) -o $@ \n");

  fclose(f);
  
  /* start the make */
  fprintf(stderr,"Check out %s/%s\n",wd,makefilename);
  sprintf(command,"make -f %s",makefilename);
  system(command);

  /* Has the executable been created */
  sprintf(buffer,"service-%s-upfed",pd->file);
  if (stat(buffer,&st))
  {
    ns_errno = NetSolveUPFError;
    return -1;
  }

  *xname = strdup(buffer);
  *xpath = strdup(wd);
  return 1;
}

#endif

