/*****************************************************************/
/*      server_init.c                                            */
/*      Henri Casanova                                           */
/*---------------------------------------------------------------*/
/*  serverInit()                                                 */
/*****************************************************************/

#include <netdb.h>
#include "core.h"
#include "serverglobal.h"
#include "netsolveutil.h"
#include "serverinit.h"
#include "workloadmanager.h"
#include "netsolvesignals.h"
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include "nwsutils.h"

int daemon_init();

/*
 *  server_init()
 *
 *  Performs all the initializations required by the server
 */
int server_init(int argc,char **argv)
{

  /* initialize the NETSOLVE_ROOT path */
  global.netsolve_root_path=getNetSolveRoot();
  if (global.netsolve_root_path == NULL)
  {
    fprintf(stderr,"The NETSOLVE_ROOT variable should be set\n");
    ns_errno = NetSolveSetNetSolveRoot;
    return -1;
  }  

  set_defaults();

  /* Initialize the NSLinked Lists */
  global.problems = newNSLinkedList();
  global.agents   = newNSLinkedList();

  check_args(argc, argv);

  /* Local IBP server ? */
  if(setLocalIBP(global.config_file) == -1)
    return -1;

#ifndef NO_OUTPUT
  fprintf(stderr,"Initializing Network connections...\n");
#endif

  /* Initialize my_self */
  if (initMySelf() == -1)
    return -1;

#ifndef NO_OUTPUT
  fprintf(stderr,"Condor Server?...\n");
#endif
  /* Condor ? */
  if (setCondor(global.config_file) == -1)
    return -1;

#ifndef NO_OUTPUT
  fprintf(stderr,"Scratch Space...\n");
#endif
  if (setScratchSpace(global.config_file) == -1)
    return -1;

#ifndef NO_OUTPUT
  fprintf(stderr,"MPI ?...\n");
#endif
  /* MPI ? */
  if (setMPI(global.config_file) == -1)
    return -1;

#ifndef NO_OUTPUT
  fprintf(stderr,"Number of processors...\n");
#endif
  /* Number of processors ? */
  if (setNumberProc(global.config_file) == -1)
    return -1;

  /* workload threshhold ? */
#ifndef NO_OUTPUT
  fprintf(stderr,"Workload Threshhold...\n");
#endif
  if (setWorkloadThreshhold(global.config_file) == -1)
    return -1;

  /* Attached storage spaces */
#ifndef NO_OUTPUT
  fprintf(stderr,"Storage devices...\n");
#endif
  if (setDisks(global.config_file) == -1)
    return -1;

  /* restrictions */
#ifndef NO_OUTPUT
  fprintf(stderr,"User Access Control...\n");
#endif
  if (setRestrictions(global.config_file) == -1)
    return -1;

/* Signals */
  if (setSignals() == -1)
    return -1;

#ifndef NO_OUTPUT
  fprintf(stderr,"Problems...\n");
#endif
  /* Problems */
  if (setProblems(global.config_file) == -1)
    return -1;
  /* Conflicts ? */
  if (areCustomizeConflicts() == -1)
    return -1;

#ifndef NO_OUTPUT
  fprintf(stderr,"Locating Agent...\n");
#endif
  /* Agent */
  if (setMaster(global.config_file) == -1)
    return -1;

#ifndef NO_OUTPUT
  fprintf(stderr,"Registering to NetSolve...\n");
#endif
  /* Register to the NetSolve system */
  if (registerToNetSolve(global.master,1) == -1)
    return -1;

#ifdef NWS
#ifdef VIEW
  fprintf(stderr,"Starting NWS Sensor...\n");
#endif

 if (setNWSSensor()==-1)
   {
   fprintf(stderr,"Impossible to start NWS Sensor\n");
   return -1;
   }
#endif

#ifndef NO_OUTPUT
  fprintf(stderr,"Starting Workload Manager...\n");
#endif
  /* Workload Manager */
  global.workload_manager_pid = startWorkloadManager();
  if (global.workload_manager_pid == -1)
  {
    fprintf(stderr,"Impossible to start workload manager\n");
    return -1;
  }

#ifdef HBMLM
#ifndef NO_OUTPUT
  fprintf(stderr,"Globus HBM Local Monitor...\n");
#endif
  /* Running the HBM local monitor */
  if (runHBMLocalMonitor() == -1)
    return -1;
#endif

#ifdef KERBEROS5
  fprintf(stderr, "Initializing Kerberos\n");
  if(krb5_init() < 0){
    return -1;
  }
#endif

  if(daemon_init() < 0 ){
    return -1;
  }

  fflush(stderr);
  fflush(stdout);
  return 0;
}

/*
 * initMySelf()
 *
 * Performs the initialization of the host descriptor
 * the IP port and the listening socket
 */
int initMySelf()
{
  /* If global.my_self->host_desc->hostname is already
   * set, it means that the user requested that we
   * bind to that IP instead of the gethostname() one.
   */
  if(!global.my_self->host_desc->hostname) {
    struct hostent *hp;
    char buffer[256];

    /* Getting my host name and IPaddr */
    if (gethostname(buffer,256) == -1)
      {
	perror("gethostname()");
	ns_errno = NetSolveSystemError;
	return -1;
      }
    if ((hp = gethostbyname(buffer)) == NULL)
      {
	perror("gethostbyname()");
	ns_errno = NetSolveSystemError;
	return -1;
      }
    BCOPY(hp->h_addr_list[0],
	  (void*)&(global.my_self->host_desc->IPaddr),sizeof(NS_IPaddr_type));
    global.my_self->host_desc->hostname = strdup(hp->h_name);
  }

  /* set up the port */
  fprintf(stderr, "Hostname set to \"%s\"\n", global.my_self->host_desc->hostname);
  global.sock = bindToFirstAvailablePort(PORT_INIT,
                                         &(global.my_self->port), 
					 global.my_self->host_desc->hostname);
  if (global.sock == -1)
  {
    fprintf(stderr,"Cannot bind to port !!\n");
    ns_errno = NetSolveSystemError;
    return -1;
  }

#ifndef NO_OUTPUT
  fprintf(stderr,"Binding on port %d...\n",global.my_self->port);
#endif

  global.my_self->host_desc->host_status = HOST_UP;
  global.my_self->host_desc->data_format = getArch();
  if (global.my_self->host_desc->data_format == -1)
  {
    fprintf(stderr,"Impossible to determine my architecture\n");
    return -1;
  }

#ifndef NO_OUTPUT
  fprintf(stderr,"Running the LINPACK benchmark...");
  fflush(stderr);
#endif
  global.my_self->host_desc->speed = kflops();

#ifndef NO_OUTPUT
  fprintf(stderr,"%d kflops\n",global.my_self->host_desc->speed);
#endif

  /* History */

  global.my_self->network_history->latency   = -1;
  global.my_self->network_history->bandwidth = -1;
  global.my_self->workload_history->w        = -1;

  /* Local IBP */
  fprintf(stderr, "Local IBP server: ");
  if(global.local_ibp) {
    struct hostent *hp;
    hp = gethostbyname(global.local_ibp);
    if(hp) {
      BCOPY(hp->h_addr_list[0], &global.my_self->IBPlocal, sizeof(NS_IPaddr_type));
      fprintf(stderr, "%lu\n", (unsigned long)global.my_self->IBPlocal);
    } else {
      fprintf(stderr, "None -> resolution for \"%s\" failed\n",
	      global.local_ibp);
      global.my_self->IBPlocal = 0;
    }
  } else {
    fprintf(stderr, "None specified\n");
    global.my_self->IBPlocal = 0;
  }


  return 0;
}

/*
 * setCondor()
 *
 * Sets the global variables in case of a condor server
 */
int setCondor(char *filename)
{
  FILE *f;
  char line[256];
  int linecount = 0;
  char *token1,*token2,*token3;

  /* Opening the file */
  f = fopen(filename,"r");

  if (f == NULL)
  {
    fprintf(stderr,"Impossible to open file '%s'\n",filename);
    return -1;
  }

  while(readLine(f,line,&linecount))
  {
    if (strncmp(line,"@CONDOR:",8))
      continue;
    fclose(f);
    token1 = (char *)strtok(line,":\t\n");
    token2 = (char *)strtok(NULL," \t");
    token3 = (char *)strtok(NULL," \t\n");
    if ((token2 == NULL)||(token3 == NULL))
    {
      fprintf(stderr,"%s,%d: Syntax error\n",filename,linecount);
      fclose(f);
      return -1;
    }
    global.Condor_path = (char *)strdup(token2);
    global.Condor_nb = atoi(token3);
    global.Condor_path[strlen(global.Condor_path)] = '\0';
    fclose(f);
    return 0;
  }
  fclose(f);
  return 0;
}

/*
 * setDisks()
 */
int setDisks(char *filename)
{
  FILE *f;
  char line[256];
  int linecount = 0;
  char *token1,*token2,*token3;

  /* Opening the file */
  f = fopen(filename,"r");
  if (f == NULL)
  {
    fprintf(stderr,"Impossible to open file '%s'\n",filename);
    return -1;
  }

  while(readLine(f,line,&linecount))
  {
    if (strncmp(line,"@DISK",5))
      continue;
    token1 = (char *)strtok(line," \t\n");
    token2 = (char *)strtok(NULL," \t\n");
    token3 = (char *)strtok(NULL," \t\n");
    if ((token2 == NULL)||(token3 == NULL))
    {
      fprintf(stderr,"%s,%d: Syntax error\n",filename,linecount);
      fclose(f);
      return -1;
    }
    (global.my_self->nb_disks)++; 
    global.my_self->disks = (char **)myRealloc(global.my_self->disks,
                        (global.my_self->nb_disks)*sizeof(char*));
    global.my_self->disks[global.my_self->nb_disks -1] = strdup(token2);
    global.my_self->diskspaces = (int *)myRealloc(global.my_self->diskspaces,
                        (global.my_self->nb_disks)*sizeof(int));
    global.my_self->diskspaces[global.my_self->nb_disks -1] = atoi(token3);
  }
  fclose(f);
  return 0;
}

/*
 * setScratchSpace()
 *
 * Sets the path to the scratch area
 */
int setScratchSpace(char *filename)
{
  FILE *f;
  char line[256];
  int linecount = 0;
  char *token1,*token2;

  /* Opening the file */
  f = fopen(filename,"r");

  if (f == NULL)
  {
    fprintf(stderr,"Impossible to open file '%s'\n",filename);
    return -1;
  }

  while(readLine(f,line,&linecount))
  {
    if (strncmp(line,"@SCRATCH:",9))
      continue;
    fclose(f);
    token1 = (char *)strtok(line,":\t\n");
    token2 = (char *)strtok(NULL," \t\n");
    if (token2 == NULL)
    {
      fprintf(stderr,"%s,%d: Syntax Error\n",filename,linecount);
      return -1;
    }
    global.scratch_path = (char *)strdup(token2);
    global.scratch_path[strlen(global.scratch_path)] = '\0';
    return 0;
  }
  return 0;
}

/*
 * setMPI()
 *
 * Sets the global variables in case of an MPI server
 *
 */
int setMPI(char *filename)
{
  FILE *f;
  char line[256];
  int linecount = 0;
  char *token1,*token2,*token3;

  /* Opening the file */
  f = fopen(filename,"r");

  if (f == NULL)
  {
    fprintf(stderr,"Impossible to open file '%s'\n",filename);
    return -1;
  }

  global.nb_MPInodes=0;
  global.MPInodefile=NULL;
  global.MPI_path=NULL;
  while(readLine(f,line,&linecount))
  {
    if (!strncmp(line,"@MPIHOSTS",7))
    {
      token1 = (char *)strtok(line," \t\n");
      token2 = (char *)strtok(NULL," \t\n");
      token3 = (char *)strtok(NULL," \t\n");
      if ((token2 == NULL)||(token3 == NULL))
      {
        fprintf(stderr,"%s,%d: Syntax Error\n",filename,linecount);
        fclose(f);
        return -1;
      }
      global.nb_MPInodes = atoi(token3);
      if (global.nb_MPInodes <= 0)
      {
        fprintf(stderr,"%s,%d: Invalid number of nodes\n",
                       filename,linecount);
        fclose(f);
        return -1;
      }
      global.MPInodefile = strdup(token2);
    }

    if (!strncmp(line,"@MPIPATH",8))
    {
      token1 = (char *)strtok(line," \t\n");
      token2 = (char *)strtok(NULL," \t\n");
      if (token2 == NULL)
      {
        fprintf(stderr,"%s,%d: Syntax Error\n",filename,linecount);
        fclose(f);
        return -1;
      }
      global.MPI_path=strdup(token2);
    }
  }
  fclose(f);
  if (((global.nb_MPInodes == 0)&&(global.MPI_path!= NULL))||
      ((global.nb_MPInodes != 0)&&(global.MPI_path== NULL)))
  {
    fprintf(stderr,"%s: MPI might be missing information (but not on some MPPs)\n",filename);
    return -1;
  }
  return 0;
}

/*
 * setNumberProc()
 *
 * Sets the number of processors available
 */
int setNumberProc(char *filename)
{
  FILE *f;
  char line[256];
  int linecount = 0;

  /* Opening the file */
  f = fopen(filename,"r");

  if (f == NULL)
  {
    fprintf(stderr,"Impossible to open file '%s'\n",filename);
    return -1;
  }

  while(readLine(f,line,&linecount))
  {
    if (strncmp(line,"@PROC:",6) == 0)
    {
      global.my_self->host_desc->number_proc = atoi(&(line[6]));
      if (global.my_self->host_desc->number_proc <= 0)
      {
        fprintf(stderr,"%s, %d: Invalid number of processors\n",
                filename,linecount);
        fclose(f);
        return -1;
      }
      fclose(f);
      return 0;
    }
  }
  fclose(f);
  return 0;
}


/*
 * Set the local IBP server (if any)
 */
int setLocalIBP(char *filename)
{
  FILE *f;
  char line[256];
  int linecount = 0;

  /* Opening the file */
  f = fopen(filename,"r");

  if (f == NULL)
  {
    fprintf(stderr,"Impossible to open file '%s'\n",filename);
    return -1;
  }

  global.local_ibp = 0;

  while(readLine(f,line,&linecount))
  {
    if (strncmp(line,"@LOCALIBP:",10) == 0)
    {
      int ofs = 10;
      char * charp;
      while(line[ofs] && line[ofs] == ' ') ofs++;
      global.local_ibp = strdup(line+ofs);
      fclose(f);
      for(charp = global.local_ibp; *charp; charp++)
	if(*charp == '\n' || *charp == '\r') {
	  *charp = 0;
	  break;
	}
      return 0;
    }
  }
  fclose(f);
  return 0;
}


/*
 * setWorkloadThreshhold()
 *
 * Sets the workload threshhold
 */
int setWorkloadThreshhold(char *filename)
{
  FILE *f;
  char line[256];
  int linecount = 0;

  /* Opening the file */
  f = fopen(filename,"r");

  if (f == NULL)
  {
    fprintf(stderr,"Impossible to open file '%s'\n",filename);
    return -1;
  }

  while(readLine(f,line,&linecount))
  {
    if (strncmp(line,"@WORKLOADMAX:",13) == 0)
    {

      global.my_self->workload_threshhold = atoi(&(line[13]));
      if (global.my_self->workload_threshhold < -1)
      {
        fprintf(stderr,"%s, %d: Invalid workload threshhold\n",
                filename,linecount);
        fclose(f);
        return -1;
      }
      fclose(f);
      return 0;
    }
  }
  fclose(f);
  return 0;
}

/*
 * setProblems()
 *
 * calls the routine generated by the code generator
 */
int setProblems(char *filename)
{
  int cc,i;
  char **filelist=NULL;
  int listlength=0;
  int linecount;
  int count=0;
  char line[256];
  FILE *f;

  /* Count the number of problem files */
  f = fopen(filename,"r");
  if (f == NULL)
  {
    fprintf(stderr,"Impossible to open file '%s'\n",filename);
    return -1;
  }

  linecount = 0;
  while(readLine(f,line,&linecount))
  {
    if (strncmp(line,"@PROBLEMS:",10))
      continue;
    while(readLine(f,line,&linecount))
    {
      if (line[0]=='@')
        break;
      else
        listlength++;
    }
  }
  fclose(f);

  /* Allocate the space for the list */
  filelist = (char **)calloc(listlength,sizeof(char*));

  /* fill the list with the file names */
  f = fopen(filename,"r");
  if (f == NULL)
  {
    fprintf(stderr,"Impossible to open file '%s'\n",filename);
    return -1;
  }
  linecount = 0;
  while(readLine(f,line,&linecount))
  {
    if (strncmp(line,"@PROBLEMS:",10))
      continue;
    while(readLine(f,line,&linecount))
    {
      if (line[0]=='@')
        break;
      line[strlen(line)-1] = '\0';
      filelist[count++] = strdup(1+strrchr(line,'/'));
    }
  }
  fclose(f);

  /* Call the generated routine */
  cc = problemInit(filelist,listlength);

  for (i=0;i<listlength;i++)
    free(filelist[i]);
  free(filelist);
  if (cc == -1)
    return -1;
  global.nb_problems = NScountList(global.problems);
  return 0;
}

/*
 * setMaster()
 *
 * Build the agent descriptor
 */
int setMaster(char *filename)
{
  FILE *f;
  char line[256];
  struct hostent *hp;
  char *token1,*token2,*token3;
  int linecount = 0;
  char *mastername=NULL;
  NS_AgentDesc *agentdesc;

  /* Opening the file */
  f = fopen(filename,"r");

  if (f == NULL)
  {
    fprintf(stderr,"Impossible to open file '%s'\n",filename);
    return -1;
  }

  while(readLine(f,line,&linecount))
  {
    if (strncmp(line,"@AGENT:",7) == 0)
    {
      token1 = (char *)strtok(line,":\t\n");
      token2 = (char *)strtok(NULL," \t\n");
      token3 = (char *)strtok(NULL," \t\n");
      if (token2 == NULL)
      {
        fprintf(stderr,"%s,%d: Syntax error\n",filename,linecount);
        return -1;
      }
      mastername = (char *)strdup(token2);
      if (token3 != NULL)
      {
        if (!strcmp(token3,"*"))
          global.my_self->broadcast = 1;
        else
          global.my_self->broadcast = 0;
      }
      else
      {
        global.my_self->broadcast = 0;
      }
    }
  }
  fclose(f);
  if (mastername == NULL)
  {
    fprintf(stderr,"%s: No agent specified\n",filename);
    return -1;
  }

  /* Get the full name and IPaddr of the agent */
  hp = gethostbyname(mastername);
  if (hp == NULL)
  {
    fprintf(stderr,"Impossible to resolve hostname '%s'\n",mastername);
    return -1;
  } 
  agentdesc = newAgentDesc();
  agentdesc->port = AGENT_PORT;
  agentdesc->host_desc = newHostDesc();
  agentdesc->host_desc->hostname= 
      strdup(hp->h_name);
  BCOPY(hp->h_addr_list[0],
       (void*)&(agentdesc->host_desc->IPaddr),
       sizeof(NS_IPaddr_type));
  free(mastername);

  addAgent(agentdesc);
  global.master = agentdesc;
  return 0;
}

/*
 * registerToNetSolve
 *
 * Apply for computation
 */
int registerToNetSolve(NS_AgentDesc *agent,int report)
{
  int sock;
  NS_Communicator *comm;
  NS_AgentDesc *ad;
  int nb_agents,i;
  int tag;

  sock = connectToSocket(agent->host_desc->hostname,
                           agent->host_desc->IPaddr,
                           agent->port);
  if (sock == -1)
  {
    netsolvePerror(__FILE__,__LINE__);
    fprintf(stderr,"Impossible to contact agent on '%s'\n",
                   agent->host_desc->hostname);
    return -1;
  }

  comm=initTransaction(sock,DATA_XDR);
  if (comm == NULL)
  {
    netsolvePerror(__FILE__,__LINE__);
    endTransaction(comm);
    return -1;
  }
  if (sendInt(comm,NS_PROT_SV_REGISTER) == -1)
  {
    netsolvePerror(__FILE__,__LINE__);
    endTransaction(comm);
    return -1;
  }
  if (sendServerDesc(comm,global.my_self) == -1)
  {
    netsolvePerror(__FILE__,__LINE__);
    endTransaction(comm);
    return -1;
  }
  if (sendAllProblems(comm) == -1)
  {
    netsolvePerror(__FILE__,__LINE__);
    endTransaction(comm);
    return -1;
  }

  if (recvInt(comm,&tag) == -1)
  {
    netsolvePerror(__FILE__,__LINE__);
    endTransaction(comm);
    return -1;
  }
  if (tag == NS_PROT_REGISTRATION_REFUSED)
  {
    fprintf(stderr,"The agent rejected my application\n");
    ns_errno = NetSolveNotAllowed;
    endTransaction(comm);
    return -1;
  }
  if (tag != NS_PROT_REGISTRATION_ACCEPTED)
  {
    endTransaction(comm);
    ns_errno = NetSolveProtocolError;
    netsolvePerror(__FILE__,__LINE__);
    return -1;
  }
#ifndef NO_OUTPUT
  fprintf(stderr,"Registration accepted\n");
#endif
  if (recvInt(comm,&nb_agents) == -1)
  {
    netsolvePerror(__FILE__,__LINE__);
    endTransaction(comm);
    return -1;
  }
  for (i=0;i<nb_agents;i++)
  {
    ad = recvAgentDesc(comm);
    if (ad == NULL)
    {
      netsolvePerror(__FILE__,__LINE__);
      endTransaction(comm);
      return -1;
    }
    if ((*lookupAgent(&(ad->host_desc->IPaddr))) == NULL)
      addAgent(ad);
    if (report)
    {
      if (registerToNetSolve(ad,0) == -1)
      {
#ifndef NO_OUTPUT
        fprintf(stderr,"WARNING:Possible incoherence in the system\n");
#endif
      }
    }
  }
  endTransaction(comm);
  return 0;
}

/*
 * sendAllProblems()
 */
int sendAllProblems(NS_Communicator *comm)
{
  NS_LinkItem *link;

  if (sendInt(comm,global.nb_problems) == -1)
    return -1;
  link = global.problems->start;
  while(link != NULL)
  {
    if (sendProblemDesc(comm,(NS_ProblemDesc *)(link->content)) == -1)
    {
      netsolvePerror(__FILE__,__LINE__);
      return -1;
    }
    link = link->next;
  }
  return 0;
}

#ifdef HBM
/*
 * runHBMLocalMonitor()
 *
 * Runs the Heart Beat Local Monitor
 */
int runHBMLocalMonitor()
{
  char commandline[256];
  char path[256];
  int pid;
  char *arch;

  /* clean up the HBM files if any */
  sprintf(commandline,"rm -rf /tmp/hbmlm*");
  system(commandline);

  global.hbmlm_pid = fork();
  if (global.hbmlm_pid == -1)
  {
    ns_errno = NetSolveSystemError;
    netsolvePerror(__FILE__,__LINE__);
    return -1;
  }

  if (global.hbmlm_pid != 0)
    return 0;
 
  arch = getNetSolveArch();
  if (execl(HBMLM,"hbdlm",(char *)0) == -1)
  {
    fprintf(stderr,"Warning: Unable to start the HBM Local Monitor.\n");
    perror("execl");
    exit(-1);
  }
  exit(0);
}
#endif

/*
 * setRestrictions()
 *
 * Set the user access restrictions
 */
int setRestrictions(char *filename)
{
  NS_RestrictionDesc *r;
  FILE *f;
  char line[256];
  int linecount = 0;
  int count = 0;
  char *token1,*token2;

  /* Count the number of problem files */
  f = fopen(filename,"r");
  if (f == NULL)
  {
    fprintf(stderr,"Impossible to open file '%s'\n",filename);
    ns_errno = NetSolveFileError;
    return -1;
  }

  linecount = 0;
  while(readLine(f,line,&linecount))
  {
    if (strncmp(line,"@RESTRICTIONS:",14))
      continue;
    while(readLine(f,line,&linecount))
    {
      if (line[0]=='@')
        break;
      else
        count++;
    }
  }
  fclose(f);

  if (count == 0)
  {
    fprintf(stderr,"%s,%d: At least one restriction must be specified.\n",
               filename, linecount);
    ns_errno = NetSolveInternalError;
    return -1;
  }

  /* Allocate the space for the list */
  global.my_self->nb_restrictions = count;
  global.my_self->restrictions = 
          (NS_RestrictionDesc **)calloc(count,sizeof(NS_RestrictionDesc*));

  /* fill the list with the file names */
  f = fopen(filename,"r");
  if (f == NULL)
  {
    fprintf(stderr,"Impossible to open file '%s'\n",filename);
    ns_errno = NetSolveFileError;
    return -1;
  }
  linecount = 0;
  count = 0;
  while(readLine(f,line,&linecount))
  {
    if (strncmp(line,"@RESTRICTIONS:",14))
      continue;
    while(readLine(f,line,&linecount))
    {
      if (line[0]=='@')
        break;
      token1 = strtok(line," \t");
      token2 = strtok(NULL," \n");
      if ((token1 == NULL)||(token2 == NULL))
      {
        fprintf(stderr,"%s,%d: Syntax Error\n",filename,linecount);
        fclose(f);
        return -1;
      }
      if (atoi(token2) <= 0)
      {
        fprintf(stderr,"%s,%d: Invalid number of requests\n",filename,linecount);
        fclose(f);
        ns_errno = NetSolveInternalError;
        return -1;
      }
      r = newRestrictionDesc();
      r->domain = strdup(token1); 
      netsolveCaps(r->domain);
      r->max_pending = atoi(token2);
      r->pending = 0;
      (global.my_self->restrictions)[count++] = r;
    }
  }
  fclose(f);
  return 0;
}

/*
 * areCustomizeConflicts()
 */
int areCustomizeConflicts()
{
  int condor;
  NS_ProblemDesc *pd;
  NS_LinkItem *item = global.problems->start;
  char *customized;

  condor =  (global.Condor_path != NULL);

  if (!condor)
    return 0;

  while(item != NULL)
  {
    pd = (NS_ProblemDesc *)(item->content);
    customized = pd->customized;
    if (strcmp(customized,"CONDOR")&&strcmp(customized,"NONE"))
    {
      fprintf(stderr,"Problem '%s': a Condor server cannot handle '%s' customization.\n", 
                       pd->nickname,customized);
      ns_errno = NetSolveInternalError;
      return -1;
    }
    free(pd->customized);
    pd->customized = strdup("CONDOR");
    item = item->next;
  }
  return 0;
}

int set_defaults(){
  char buffer[256];

  global.my_self = newServerDesc();
  global.my_self->host_desc = newHostDesc();
  global.my_self->workload_history = newWorkloadHistory();
  global.my_self->network_history = newNetworkHistory();

  /*** set logfile to $NETSOLVE_ROOT/SERVER_LOGFILE **/
  sprintf(buffer, "_%s", getHostName());
  global.log_file=(char *)calloc(strlen(global.netsolve_root_path) +
                                   strlen(SERVER_LOGFILE) + 2 + strlen(buffer),
				 sizeof(char));
  strcpy(global.log_file, global.netsolve_root_path);
  strcat(global.log_file, "/");
  strcat(global.log_file, SERVER_LOGFILE);
  strcat(global.log_file, buffer);

  global.master = NULL;

  sprintf(buffer,"%s/server_config",global.netsolve_root_path);
  global.config_file = strdup(buffer);
  
  global.Condor_path = NULL;
  global.Condor_nb = 0;

  global.scratch_path = (char *)strdup("/tmp/");

  global.my_self->host_desc->number_proc = 1;

  global.my_self->workload_threshhold = -1;
#ifdef KERBEROS5
  global.require_auth = 0;
#endif
  return 0;
}

int check_args(int argc, char **argv){
  int i;

  for(i=1; i<argc; i++){
    if(strcmp(argv[i], "-f") == 0){
      if(i == argc-1){   /* last argument? */
        print_usage(argv[0]);
        return -1;
      }
      else{
        free(global.config_file);
        global.config_file = strdup(argv[i+1]);
        i++;
      }
    }

    else if(strcmp(argv[i], "-h") == 0) {
      if(i == argc-1) {
	print_usage(argv[0]);
	return -1;
      } else {
	global.my_self->host_desc->hostname = strdup(argv[i+1]);
	i++;
      }
    }

    else if(strcmp(argv[i], "-p") == 0) {
      if(i == argc-1) {
	print_usage(argv[0]);
	return -1;
      } else {
	global.my_self->port = strtoul(argv[i+1], 0, 0);
	i++;
	if(global.my_self->port == ULONG_MAX) { /* strtoul() returns ULONG_MAX on error */
	  fprintf(stderr, "Hey, port number given must be a valid port number!\n");
	  return -1;
	}
      }
    }

    else if(strcmp(argv[i], "-l") == 0){
      if(i == argc-1){   /* last argument? */
        print_usage(argv[0]);
        return -1;
      }
      else{
        free(global.log_file);
        global.log_file = (char *)strdup(argv[i+1]);
        i++;
      }
    }
    else if(strcmp(argv[i], "-k") == 0){
#ifdef KERBEROS5
      global.require_auth = 1;
#else
      fprintf(stderr, "-k option only useful when compiling with KERBEROS libraries\n");
      return -1;
#endif
    }

    else{
      print_usage(argv[0]);
      return -1;
    }
  }

  return 0;
}

void print_usage(char *command){
  fprintf(stderr,"Usage: %s [ -f config_file ] [ -k ] [ -l logfile ] [ -p port# ]\n"
                  , command);
  exit(-1);
}

int daemon_init(){
  pid_t pid;
  int fd;

  pid = fork();

  if(pid < 0){
    perror("fork()");
    return -1;
  }

  if(pid != 0) /* parent exits */
    exit(0);

  setsid();   /* become session leader */

  pid = fork();

  if(pid < 0){
    perror("fork()");
    return -1;
  }

  if(pid != 0) /* 1st child exits */
    exit(0);

  chdir(global.netsolve_root_path);
  umask(0);

  fd = open("/dev/null", O_RDWR|O_CREAT);
  if(fd < 0){
    perror("open()");
    return -1;
  }
  close(0);
  if(dup2(fd, 0) < 0){
    perror("dup2()");
    return -1;
  }
  close(fd);

#ifndef NO_OUTPUT
  fprintf(stderr, "\n-- Server giving up terminal control! --\n\n");
  fprintf(stderr, "All further terminal output will be sent to the"
                  " log file:\n\t\"%s\"\n\n", global.log_file);
#endif

  remove(global.log_file);
  fd = open(global.log_file, O_RDWR|O_CREAT, 0644);
  if(fd < 0){
    perror("open()");
    return -1;
  }
  close(1);
  if(dup2(fd, 1) < 0){
    perror("dup2()");
    return -1;
  }
  close(2);
  if(dup2(fd, 2) < 0){
    perror("dup2()");
    return -1;
  }
  close(fd);

  return 0;
}

int setNWSSensor()
{
int pid;
pid = startNWSSensor(NWS_DEFAULT_SENSOR_PORT,global.my_self->host_desc->hostname, global.master->host_desc->hostname,NWS_DEFAULT_NAMESERVER_PORT,global.master->host_desc->hostname,NWS_DEFAULT_MEMORY_PORT);
return pid;
}

#ifdef KERBEROS5
int krb5_init(){
  krb5_error_code retval;

  if (global.require_auth)
  {
    retval = krb5_init_context (&(global.context));

    if (retval)
    {
      fprintf (stderr,
           "krb5_init_context failed: %s\n", error_message(retval));
      return -1;
    }

    global.keytab = NULL;   /* Maybe do something seperate later */

    if (getenv ("NETSOLVE_KEYTAB"))
    {
      retval = krb5_kt_resolve (global.context, getenv ("NETSOLVE_KEYTAB"),
                &(global.keytab));
    }
    else
    {
      fprintf (stderr, "NETSOLVE_KEYTAB not defined\n");
      return -1;
    }
    if (!getenv ("NETSOLVE_USERS"))
    {
      fprintf (stderr, "NETSOLVE_KEYTAB not defined\n");
      return -1;
    }
    if (retval)
    {
      fprintf (stderr, "krb5_kt_resolve failed: %s\n",
           error_message (retval));
      return -1;
    }

    retval = krb5_sname_to_principal (global.context, NULL, "netsolve",
                      KRB5_NT_SRV_HST, &(global.server));

    if (retval)
    {
      fprintf (stderr, "krb5_sname_to_principal failed: %s\n",
           error_message (retval));
      return -1;
    }
  }
}
#endif /* KERBEROS5 */
  
