/*****************************************************************/
/*      server_process_message.c                                 */
/*      Henri Casanova                                           */
/*---------------------------------------------------------------*/
/*  processMessage()                                             */
/*****************************************************************/

#include "core.h"
#include "serverglobal.h"
#include "serverprocessmessage.h"
#include "workloadmanager.h"
#include "pong.h"
#include "generateservice.h"
#include "abort.h"
#include <sys/types.h>
#include <signal.h>
#include <unistd.h>
#include <time.h>

/*
 *  processMessage()
 *
 *  A message has been received. Let's see what it's about
 */
void processMessage(int sock)
{
  NS_Communicator *comm;
  int tag;

  comm = acceptTransaction(sock);
  if (comm == NULL)
  {
    netsolvePerror(__FILE__,__LINE__);
    return;  
  }
  
  if (recvInt(comm,&tag) == -1)
  {
    netsolvePerror(__FILE__,__LINE__);
    endTransaction(comm);
    return;
  }

  switch(tag)
  {
    case NS_PROT_ARE_YOU_THERE:
      processAreYouThere(comm);
      endTransaction(comm);
      break;
    case NS_PROT_NEW_AGENT: /* No fork */
      processNewAgent(comm);  
      endTransaction(comm);
      break;
    case NS_PROT_PONG_REQUEST: /* fork */
      if (fork())
      {
        endTransaction(comm);
        return;
      }
      else
      {
        close(global.sock);
        processPongRequest(comm);
        endTransaction(comm);
        exit(0);
      }
      break;
    case NS_PROT_MULTINODE_SPAWN:
      processMultinodeSpawn(comm);
      endTransaction(comm);
      break;
    case NS_PROT_PROBLEM_SOLVE:
      processProblemSolve(comm);
      endTransaction(comm);
      break;
    case NS_PROT_PROBLEM_SOLVE_SEQUENCE:
      processProblemSolve_sequence(comm);
      endTransaction(comm);
      break;
    case NS_PROT_REGISTER_AGAIN:
      processRegisterAgain(comm);
      endTransaction(comm);
      break;
    case NS_PROT_KILL_SERVER:
      processKillServer(comm);
      endTransaction(comm);
      break;
    case NS_PROT_SERVICE_FINISHED:
      processServiceFinished(comm);
      endTransaction(comm);
      break;
    case NS_PROT_KILL_JOB:
      processKillJob(comm);
      endTransaction(comm);
      break;
    default:
      if (sendInt(comm,NS_PROT_PROTOCOL_ERROR) == -1)
        netsolvePerror(__FILE__,__LINE__);
      endTransaction(comm);
      return;
  }
}

/*
 * processKillJob()
 */
void processKillJob(NS_Communicator *comm)
{
  int pid;
  int status;

  if (recvInt(comm,&pid) == -1)
  {
    netsolvePerror(__FILE__,__LINE__);
    return;
  }
  if (kill(pid,SIGKILL) != 0)
  {
#ifdef VIEW
    perror("kill");
#endif
    status = -1;
  }
  else 
    status = 1;

  if (status == 1)
  { 
    if (sendInt(comm,NetSolveOK) == -1)
    {
      netsolvePerror(__FILE__,__LINE__);
      return;
    }
  }
  else
  {
    if (sendInt(comm,NetSolveServerError) == -1)
    {
      netsolvePerror(__FILE__,__LINE__);
      return;
    }
  }
  return;
}

/*
 * processAreYouThere()  
 */
void processAreYouThere(NS_Communicator *comm)  
{
  if (sendInt(comm,NS_PROT_I_AM_THERE) == -1)
  {
    netsolvePerror(__FILE__,__LINE__);
    return;
  }
  return;
}

/*  
 * processKillServer()
 */   
void processKillServer(NS_Communicator *comm)
{   
  char *username = getUserName();
  char *s;
    
  if (recvString(comm,&s) == -1)
  {
    netsolvePerror(__FILE__,__LINE__);
    return;
  }

  if (username == NULL)
  {   
#ifdef VIEW
    fprintf(stderr,"Killable by anyone.\n");
#endif
    username = s;
  }   
      
  if (strcmp(s,username))
  {
    if (sendInt(comm,NS_PROT_NOT_ALLOWED) == -1)
      netsolvePerror(__FILE__,__LINE__);
    free(username);
    return;
  }
  else
  {  
    if (sendInt(comm,NS_PROT_KILLED) == -1)
      netsolvePerror(__FILE__,__LINE__);
#ifdef VIEW
    fprintf(stderr,"Murdered\n");
#endif
    Abort();
  } 
  free(username);
}   

/*
 * processRegisterAgain()
 */
void processRegisterAgain(NS_Communicator *comm)
{
  NS_IPaddr_type IPaddr;
  NS_AgentDesc **alist;

  if (recvIPaddr(comm,&IPaddr) == -1)
  {
    netsolvePerror(__FILE__,__LINE__);
    return;
  }
  alist = lookupAgent(&IPaddr);
  if (*alist == NULL)
  {
    free(alist);
    return;
  }
  registerToNetSolve(*alist,0);
  free(alist);
  return;

}


/*
 * processNewAgent()
 */
void processNewAgent(NS_Communicator *comm)
{
  NS_AgentDesc *ad;

  ad = recvAgentDesc(comm);
  if (ad == NULL)
  {
    netsolvePerror(__FILE__,__LINE__);
    return;
  }

  addAgent(ad); 

  /* Restart the workload manager */
  (void)kill(global.workload_manager_pid,SIGTERM);
  global.workload_manager_pid = startWorkloadManager();

  return;
}

/*
 * processPongRequest()
 *
 * This routine asks for the server to perform a pong measurement
 * between himself and some other host
 */
void processPongRequest(NS_Communicator *comm)
{
  NS_HostDesc *hd;
  int port;
  int latency,bandwidth;
  time_t date;
  int idate;

  hd = recvHostDesc(comm);
  if (hd == NULL)
  {
    netsolvePerror(__FILE__,__LINE__);
    return;
  }
  if (recvInt(comm,&port) == -1)
  {
    netsolvePerror(__FILE__,__LINE__);
    free(hd);
    return; 
  }

  time(&date);
  doPong(hd->hostname,hd->IPaddr,port,&latency,&bandwidth);

  if (sendInt(comm,latency) == -1)
  {
    netsolvePerror(__FILE__,__LINE__);
    free(hd);
    return;
  }
  if (sendInt(comm,bandwidth) == -1)
  {
    netsolvePerror(__FILE__,__LINE__);
    free(hd);
    return;
  }
  idate = (int)date;
  if (sendInt(comm,idate) == -1)
  {
    netsolvePerror(__FILE__,__LINE__);
    free(hd);
    return;
  }
  free(hd);
  return;
}

/*
 * processPongBandwidth()
 */
void processPongBandwidth(NS_Communicator *comm)
{
  char *data;
  int size;

  if (recvInt(comm,&size) == -1)
  {
    netsolvePerror(__FILE__,__LINE__);
    return;
  }
  data = (char *)calloc(size,sizeof(char));
  if (recvArray(comm,NETSOLVE_CHAR,data,size) == -1)
  {
    netsolvePerror(__FILE__,__LINE__);
    return;
  }
  if (sendArray(comm,NETSOLVE_CHAR,data,size) == -1)
  {
    netsolvePerror(__FILE__,__LINE__);
    return;
  }
  return;
}

/*
 * Receive a single node, and spawn it.  Data
 * are already in IBP.
 */
void processMultinodeSpawn(NS_Communicator * comm)
{
  NS_ProblemDesc * pd;
  NS_ProblemDesc **plist;
  int restriction_index;
  int cpid;

  fprintf(stderr, "Got a multinode spawn request!\n");

  sendInt(comm, NS_PROT_OK);

  pd = recvProblemDesc(comm);
  if(!pd) {
    netsolvePerror(__FILE__,__LINE__);
    return;
  }

  fprintf(stderr, "In server_process_message: processMultinodeSpawn()\n");
  if(pd->input_names) {
    fprintf(stderr, "We HAVE input names\n");
  } else {
    fprintf(stderr, "WE DON'T HAVE INPUT NAMES!!\n");
  }

  /*
   * Ok - sanity check - see if actually know the problem
   */
  plist = lookupProblem(pd->nickname);
  if(!*plist) {
    if(sendInt(comm,NS_PROT_PROBLEM_NOT_FOUND) == -1) {
      netsolvePerror(__FILE__,__LINE__);
      freeProblemDesc(pd);
      return;
    }
    free(plist);
    return;
  }
  /*
   * And if we know the problem - check that we agree with
   * the idea of what the problem is...
   */
  if(!compareProblemDesc(pd,*plist)) {
    if(sendInt(comm,NS_PROT_BAD_SPECIFICATION) == -1) {
      netsolvePerror(__FILE__,__LINE__);
      freeProblemDesc(pd);
      return;
    }
    free(plist);
    freeProblemDesc(pd);
    return;
  }
/*    (*plist)->job_id = pd->job_id; */
/*    freeProblemDesc(pd); */
/*    pd = *plist; */
  free(plist);

  /* 
   * Check Workload Threshhold
   */
  if(global.my_self->workload_threshhold != -1) {
    if (global.my_self->workload_threshhold < getMyWorkload()) {
      fprintf(stderr,"Multinode Spawn refused: Workload beyond threshhold\n");
      if(sendInt(comm,NS_PROT_NOT_ALLOWED) == -1) {
        netsolvePerror(__FILE__,__LINE__);
        return;
      }
      return;
    }
  }

  /* 
   * Check restrictions 
   */
  restriction_index = checkRestriction(comm,global.my_self->restrictions,
				       global.my_self->nb_restrictions);
  if(restriction_index == -1) {
    if(ns_errno == NetSolveNotAllowed) {
      fprintf(stderr,"Job refused: Not an allowed client at this time\n");
      if(sendInt(comm,NS_PROT_NOT_ALLOWED) == -1) {
        netsolvePerror(__FILE__,__LINE__);
        return;
      }
      return;
    }
  }

  /*
   * At this point, either we had a system Error or
   * we found the restirction slot 
   */

  /*
   * Spawn !
   */
  cpid = fork();
  
  if(cpid < 0) {
    perror("fork(): ");
    fprintf(stderr, "Couldn't fork at %s line %i\n", __FILE__, __LINE__);
    /*
     * Send back some error so that our peer knows we're not doing this
     */
    if(sendInt(comm,NS_PROT_NOT_ALLOWED) == -1) {
      netsolvePerror(__FILE__,__LINE__);
      return;
    }
    return;
  }

  if(cpid == 0) {
    /*
     * Ok - this is a service process.  It will send back
     * the exit status of the solver to the peer that
     * requested the multinode spawn.
     */
    int wpid;
    int status;

    /*
     * Client process forked off...  Let's tell the peer
     * that it's ok.
     */  
    if(sendInt(comm,NS_PROT_OK) == -1) {
      netsolvePerror(__FILE__,__LINE__);
      exit(0);
    }

    wpid = fork();
    if(wpid < 0) {
      /* 
       * send back error 
       */     
      if(sendInt(comm,-1) == -1) {
	netsolvePerror(__FILE__,__LINE__);
	exit(0);
      }
      exit(0);
    }

    if(wpid == 0) {
      /*
       * Run service
       */
      struct in_addr adr;
      adr.s_addr = global.my_self->IBPlocal;
      
      fprintf(stderr, "ALERT: t=%li job %i is START\n",
	      time(0), pd->job_id);

      fprintf(stderr, "Will generate service process with IBP at %s\n",
	      inet_ntoa(adr));
      generateServiceProcess(0, restriction_index,
			     pd, 0, 0, 0, inet_ntoa(adr));
      exit(0);
    }

    /*
     * Ok - we're the parent then.
     */
    fprintf(stderr, "Awaiting completion of solver...\n");
    waitpid(wpid, &status, 0);
    fprintf(stderr, "Solver completed ! \n");
    fprintf(stderr, "CONFIRM: t=%li job %i is DONE\n", time(0), pd->job_id);
    
    /*
     * Send back the wait_status - and disconnect.
     */
    if(sendInt(comm,status) == -1) {
      netsolvePerror(__FILE__,__LINE__);
      exit(0);
    }

    endTransaction(comm);
    exit(0);
  }
  
  /*
   * Just return - we're done.
   */
}

/*
 * processProblemSolve()
 */
void processProblemSolve(NS_Communicator *comm)
{
  NS_ProblemDesc *pd;
  NS_ProblemDesc **plist;
  int tag,restriction_index;
  NS_IPaddr_type proxy_IPaddr;
  int proxy_port;
  char *agent;

  tag = NS_PROT_OK;
  if (sendInt (comm, tag) == -1)
  {
    netsolvePerror (__FILE__,__LINE__);
    return;
  }

  pd = recvProblemDesc(comm);
  fprintf(stderr, "server_process_message got job_id = %i\n", pd->job_id);
  if (pd == NULL)
  {
    netsolvePerror(__FILE__,__LINE__);
    return;
  }
  if ((recvString(comm,&agent) == -1) ||
      (recvIPaddr(comm,&proxy_IPaddr) == -1) ||
      (recvInt(comm,&proxy_port) == -1))
  {
    netsolvePerror(__FILE__,__LINE__);
    return;
  }
  
  plist = lookupProblem(pd->nickname);
  if (*plist == NULL)
  {
    if (sendInt(comm,NS_PROT_PROBLEM_NOT_FOUND) == -1)
    {
      netsolvePerror(__FILE__,__LINE__);
      freeProblemDesc(pd);
      return;
    }
    free(plist);
    return;
  }
  if (!compareProblemDesc(pd,*plist))
  {
    if (sendInt(comm,NS_PROT_BAD_SPECIFICATION) == -1)
    {
      netsolvePerror(__FILE__,__LINE__);
      freeProblemDesc(pd);
      return;
    }
    free(plist);
    freeProblemDesc(pd);
    return;
  }
  (*plist)->job_id = pd->job_id;
  freeProblemDesc(pd);
  pd = *plist;
  free(plist);

  /* Workload Threshhold */
  if (global.my_self->workload_threshhold != -1)
  {
    if (global.my_self->workload_threshhold < getMyWorkload())
    {
#ifdef VIEW
      fprintf(stderr,"Job refused: Workload beyond threshhold\n");
#endif
      if (sendInt(comm,NS_PROT_NOT_ALLOWED) == -1)
      {
        netsolvePerror(__FILE__,__LINE__);
        return;
      }
      return;
    }
  }

  /* Restrictions */
  restriction_index = checkRestriction(comm,global.my_self->restrictions,
                  global.my_self->nb_restrictions);
  if (restriction_index == -1)
  {
    if (ns_errno == NetSolveNotAllowed)
    {
#ifdef VIEW
      fprintf(stderr,"Job refused: Not an allowed client at this time\n");
#endif
      if (sendInt(comm,NS_PROT_NOT_ALLOWED) == -1)
      {
        netsolvePerror(__FILE__,__LINE__);
        return;
      }
      return;
    }
    if (ns_errno == NetSolveSystemError)
    {
        /* nothing */
    }
  }

  /* At this point, either we had a system Error or
     we found the restirction slot */

  if (sendInt(comm,NS_PROT_ACCEPTED) == -1)
  {
    netsolvePerror(__FILE__,__LINE__);
    return;
  }

  generateServiceProcess(comm,restriction_index,pd,agent,
                         proxy_IPaddr,proxy_port,"");
  return;
}

/*
 * processServiceFinished()
 */
void processServiceFinished(NS_Communicator *comm)
{
  int restriction_index;

  if (recvInt(comm,&restriction_index) == -1)
  {
    netsolvePerror(__FILE__,__LINE__);
    return;
  }
  if (sendInt(comm,666) == -1)
  {
    netsolvePerror(__FILE__,__LINE__);
    return;
  }

  if (restriction_index == -1)
    return;

  ((global.my_self->restrictions)[restriction_index]->pending)--;
  if ((global.my_self->restrictions)[restriction_index]->pending <= 0)
    (global.my_self->restrictions)[restriction_index]->pending = 0;
  return;
}
